# Copyright (c) 2021, Andrew Pantuso (@ajpantuso) <ajpantuso@gmail.com>
# Copyright (c) 2018, Dag Wieers (@dagwieers) <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import csv
import typing as t
from io import StringIO

from ansible.module_utils.common.text.converters import to_native

if t.TYPE_CHECKING:
    from collections.abc import Sequence

    class DialectParamsOrNone(t.TypedDict):
        delimiter: t.NotRequired[str | None]
        doublequote: t.NotRequired[bool | None]
        escapechar: t.NotRequired[str | None]
        lineterminator: t.NotRequired[str | None]
        quotechar: t.NotRequired[str | None]
        quoting: t.NotRequired[int | None]
        skipinitialspace: t.NotRequired[bool | None]
        strict: t.NotRequired[bool | None]


class CustomDialectFailureError(Exception):
    pass


class DialectNotAvailableError(Exception):
    pass


CSVError = csv.Error


def initialize_dialect(dialect: str, **kwargs: t.Unpack[DialectParamsOrNone]) -> str:
    # Add Unix dialect from Python 3
    class unix_dialect(csv.Dialect):
        """Describe the usual properties of Unix-generated CSV files."""

        delimiter = ","
        quotechar = '"'
        doublequote = True
        skipinitialspace = False
        lineterminator = "\n"
        quoting = csv.QUOTE_ALL

    csv.register_dialect("unix", unix_dialect)

    if dialect not in csv.list_dialects():
        raise DialectNotAvailableError(f"Dialect '{dialect}' is not supported by your version of python.")

    # Create a dictionary from only set options
    dialect_params = {k: v for k, v in kwargs.items() if v is not None}
    if dialect_params:
        try:
            csv.register_dialect("custom", dialect, **dialect_params)  # type: ignore
        except TypeError as e:
            raise CustomDialectFailureError(f"Unable to create custom dialect: {e}") from e
        dialect = "custom"

    return dialect


def read_csv(data: str, dialect: str, fieldnames: Sequence[str] | None = None) -> csv.DictReader:
    BOM = "\ufeff"
    data = to_native(data, errors="surrogate_or_strict")
    if data.startswith(BOM):
        data = data[len(BOM) :]

    fake_fh = StringIO(data)

    reader = csv.DictReader(fake_fh, fieldnames=fieldnames, dialect=dialect)

    return reader



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_sase_sites_attach
short_description: Resource module for organizations _sase _sites _attach
description:
  - Manage operation create of the resource organizations _sase _sites _attach. -
    > Attach sites in this organization to Secure Access. For an organization, a maximum
    of 2500 sites can be attached if they are in spoke mode or a maximum of 10 sites
    can be attached in hub mode.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  callback:
    description: Details for the callback. Please include either an httpServerId OR
      url and sharedSecret.
    suboptions:
      httpServer:
        description: The webhook receiver used for the callback webhook.
        suboptions:
          id:
            description: The webhook receiver ID that will receive information. If
              specifying this, please leave the url and sharedSecret fields blank.
            type: str
        type: dict
      payloadTemplate:
        description: The payload template of the webhook used for the callback.
        suboptions:
          id:
            description: The ID of the payload template. Defaults to 'wpt_00005' for
              the Callback (included) template.
            type: str
        type: dict
      sharedSecret:
        description: A shared secret that will be included in the requests sent to
          the callback URL. It can be used to verify that the request was sent by
          Meraki. If using this field, please also specify an url.
        type: str
      url:
        description: The callback URL for the webhook target. If using this field,
          please also specify a sharedSecret.
        type: str
    type: dict
  items:
    description: List of Meraki SD-WAN sites with the associated regions to be attached.
    elements: dict
    suboptions:
      network:
        description: Meraki SD-WAN Network to be attached.
        suboptions:
          id:
            description: Network ID of the site to attach.
            type: str
        type: dict
      region:
        description: Region metadata for the site.
        suboptions:
          slug:
            description: Unique identifying slug for the region to attach the site
              in.
            type: str
        type: dict
    type: list
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations attachOrganizationSaseSites
    description: Complete reference of the attachOrganizationSaseSites API.
    link: https://developer.cisco.com/meraki/api-v1/#!attach-organization-sase-sites
notes:
  - SDK Method used are
    organizations.Organizations.attach_organization_sase_sites,
  - Paths used are
    post /organizations/{organizationId}/sase/sites/attach,
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_sase_sites_attach:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    callback:
      httpServer:
        id: aHR0cHM6Ly93d3cuZXhhbXBsZS5jb20vd2ViaG9va3M=
      payloadTemplate:
        id: wpt_2100
      sharedSecret: secret
      url: https://webhook.site/28efa24e-f830-4d9f-a12b-fbb9e5035031
    items:
      - network:
          id: N_1234567890
        region:
          slug: us-west-1
    organizationId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "items": [
        {
          "pipelineId": "string",
          "operation": {
            "id": "string"
          },
          "status": "string",
          "counts": {
            "jobs": {
              "total": 0,
              "byStatus": {
                "completed": 0,
                "failed": 0,
                "pending": 0
              }
            }
          }
        }
      ],
      "meta": {
        "counts": {
          "items": {
            "total": 0,
            "remaining": 0
          }
        }
      }
    }
"""



# Copyright (c) 2017-2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    from collections.abc import Collection, Mapping, Sequence  # pragma: no cover

    from ._conversion.converter import RecordConverter  # pragma: no cover


def format_ttl(ttl: int | None) -> str:
    if ttl is None:
        return "default"
    sec = ttl % 60
    ttl //= 60
    mins = ttl % 60
    ttl //= 60
    h = ttl
    result = []
    if h:
        result.append(f"{h}h")
    if mins:
        result.append(f"{mins}m")
    if sec:
        result.append(f"{sec}s")
    return " ".join(result)


RecordIDT = t.TypeVar("RecordIDT")
RecordIDT_co = t.TypeVar("RecordIDT_co", covariant=True)


class IDNSRecord(t.Protocol, t.Generic[RecordIDT_co]):
    @property
    def id(self) -> RecordIDT_co: ...

    @property
    def type(self) -> str: ...

    @property
    def prefix(self) -> str | None: ...

    @property
    def target(self) -> str: ...

    @property
    def ttl(self) -> int | None: ...

    @property
    def extra(self) -> Mapping[str, str]: ...


class DNSRecord(t.Generic[RecordIDT]):
    def __init__(self, *, record_id: RecordIDT, record_type: str, target: str) -> None:
        self.id = record_id
        self.type = record_type
        self.prefix: str | None = None
        self.target = target
        self.ttl: int | None = 86400  # 24 * 60 * 60
        self.extra: dict[str, str] = {}

    def clone(self) -> DNSRecord[RecordIDT]:
        result = DNSRecord(record_id=self.id, record_type=self.type, target=self.target)
        result.prefix = self.prefix
        result.ttl = self.ttl
        result.extra = dict(self.extra)
        return result

    def __str__(self) -> str:
        data = []
        if self.id is not None:
            data.append(f"id: {self.id}")
        data.append(f"type: {self.type}")
        if self.prefix is not None:
            data.append(f"prefix: {self.prefix!r}")
        else:
            data.append("prefix: (none)")
        data.append(f"target: {self.target!r}")
        data.append(f"ttl: {format_ttl(self.ttl)}")
        if self.extra:
            data.append(f"extra: {self.extra}")
        return "DNSRecord(" + ", ".join(data) + ")"

    def __repr__(self) -> str:
        return self.__str__()


@t.overload
def sorted_ttls(ttls: Collection[int]) -> list[int]: ...


@t.overload
def sorted_ttls(ttls: Collection[int | None]) -> list[int | None]: ...


def sorted_ttls(ttls: Collection[int | None]) -> list[int | None] | list[int]:
    return sorted(ttls, key=lambda ttl: 0 if ttl is None else ttl)


def format_records_for_output(
    records: Sequence[IDNSRecord[RecordIDT_co]],
    record_name: str | None,
    prefix: str | None = None,
    record_converter: RecordConverter | None = None,
) -> dict[str, t.Any]:
    ttls = sorted_ttls({record.ttl for record in records})
    entry: dict[str, t.Any] = {
        "prefix": prefix or "",
        "type": min(record.type for record in records) if records else None,
        "ttl": ttls[0] if len(ttls) > 0 else None,
        "value": [record.target for record in records],
    }
    if record_converter:
        entry["value"] = record_converter.process_values_to_user(
            entry["type"], entry["value"]
        )
    if record_name is not None:
        entry["record"] = record_name
    if len(ttls) > 1:
        entry["ttls"] = ttls
    return entry


def format_record_for_output(
    record: IDNSRecord[RecordIDT_co],
    record_name: str | None,
    prefix: str | None = None,
    record_converter: RecordConverter | None = None,
) -> dict[str, t.Any]:
    entry: dict[str, t.Any] = {
        "prefix": prefix or "",
        "type": record.type,
        "ttl": record.ttl,
        "value": record.target,
        "extra": record.extra,
    }
    if record_converter:
        entry["value"] = record_converter.process_value_to_user(
            entry["type"], entry["value"]
        )
    if record_name is not None:
        entry["record"] = record_name
    return entry



#!/usr/bin/python

# Copyright (c) 2013, Andrew Dunham <andrew@du.nham.ca>
# Copyright (c) 2013, Daniel Jaouen <dcj24@cornell.edu>
# Copyright (c) 2015, Indrajit Raychaudhuri <irc+code@indrajit.com>
# Copyright (c) 2024, Kit Ham <kitizz.devside@gmail.com>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: homebrew_services
author:
  - "Kit Ham (@kitizz)"
requirements:
  - homebrew must already be installed on the target system
short_description: Services manager for Homebrew
version_added: 9.3.0
description:
  - Manages daemons and services using Homebrew.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    description:
      - An installed homebrew package whose service is to be updated.
    aliases: ['formula']
    type: str
    required: true
  path:
    description:
      - A V(:) separated list of paths to search for C(brew) executable. Since a package (I(formula) in homebrew parlance)
        location is prefixed relative to the actual path of C(brew) command, providing an alternative C(brew) path enables
        managing different set of packages in an alternative location in the system.
    default: '/usr/local/bin:/opt/homebrew/bin:/home/linuxbrew/.linuxbrew/bin'
    type: path
  state:
    description:
      - State of the package's service.
    choices: ['present', 'absent', 'restarted']
    default: present
    type: str
"""

EXAMPLES = r"""
- name: Install foo package
  community.general.homebrew:
    name: foo
    state: present

- name: Start the foo service (equivalent to `brew services start foo`)
  community.general.homebrew_services:
    name: foo
    state: present

- name: Restart the foo service (equivalent to `brew services restart foo`)
  community.general.homebrew_services:
    name: foo
    state: restarted

- name: Remove the foo service (equivalent to `brew services stop foo`)
  community.general.homebrew_services:
    name: foo
    state: absent
"""

RETURN = r"""
pid:
  description:
    - If the service is now running, this is the PID of the service, otherwise -1.
  returned: success
  type: int
  sample: 1234
running:
  description:
    - Whether the service is running after running this command.
  returned: success
  type: bool
  sample: true
"""

import json
import typing as t
from dataclasses import dataclass

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._homebrew import (
    HomebrewValidate,
    parse_brew_path,
)


@dataclass
class HomebrewServiceArgs:
    name: str
    state: str
    brew_path: str


@dataclass
class HomebrewServiceState:
    running: bool
    pid: int | None


def _brew_service_state(args: HomebrewServiceArgs, module: AnsibleModule) -> HomebrewServiceState:
    cmd = [args.brew_path, "services", "info", args.name, "--json"]
    rc, stdout, stderr = module.run_command(cmd, check_rc=True)

    try:
        data = json.loads(stdout)[0]
    except json.JSONDecodeError:
        module.fail_json(msg=f"Failed to parse JSON output:\n{stdout}")

    return HomebrewServiceState(running=data["status"] == "started", pid=data["pid"])


def _exit_with_state(
    args: HomebrewServiceArgs, module: AnsibleModule, changed: bool = False, message: str | None = None
) -> None:
    state = _brew_service_state(args, module)
    if message is None:
        message = f"Running: {state.running}, Changed: {changed}, PID: {state.pid}"
    module.exit_json(msg=message, pid=state.pid, running=state.running, changed=changed)


def validate_and_load_arguments(module: AnsibleModule) -> HomebrewServiceArgs:
    """Reuse the Homebrew module's validation logic to validate these arguments."""
    package: str = module.params["name"]
    if not HomebrewValidate.valid_package(package):
        module.fail_json(msg=f"Invalid package name: {package}")

    state: t.Literal["present", "absent", "restarted"] = module.params["state"]
    brew_path = parse_brew_path(module)

    return HomebrewServiceArgs(name=package, state=state, brew_path=brew_path)


def start_service(args: HomebrewServiceArgs, module: AnsibleModule) -> None:
    """Start the requested brew service if it is not already running."""
    state = _brew_service_state(args, module)
    if state.running:
        # Nothing to do, return early.
        _exit_with_state(args, module, changed=False, message="Service already running")

    if module.check_mode:
        _exit_with_state(args, module, changed=True, message="Service would be started")

    start_cmd = [args.brew_path, "services", "start", args.name]
    module.run_command(start_cmd, check_rc=True)

    _exit_with_state(args, module, changed=True)


def stop_service(args: HomebrewServiceArgs, module: AnsibleModule) -> None:
    """Stop the requested brew service if it is running."""
    state = _brew_service_state(args, module)
    if not state.running:
        # Nothing to do, return early.
        _exit_with_state(args, module, changed=False, message="Service already stopped")

    if module.check_mode:
        _exit_with_state(args, module, changed=True, message="Service would be stopped")

    stop_cmd = [args.brew_path, "services", "stop", args.name]
    module.run_command(stop_cmd, check_rc=True)

    _exit_with_state(args, module, changed=True)


def restart_service(args: HomebrewServiceArgs, module: AnsibleModule) -> None:
    """Restart the requested brew service. This always results in a change."""
    if module.check_mode:
        _exit_with_state(args, module, changed=True, message="Service would be restarted")

    restart_cmd = [args.brew_path, "services", "restart", args.name]
    module.run_command(restart_cmd, check_rc=True)

    _exit_with_state(args, module, changed=True)


def main() -> None:
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(
                aliases=["formula"],
                required=True,
                type="str",
            ),
            state=dict(
                choices=["present", "absent", "restarted"],
                default="present",
            ),
            path=dict(
                default="/usr/local/bin:/opt/homebrew/bin:/home/linuxbrew/.linuxbrew/bin",
                type="path",
            ),
        ),
        supports_check_mode=True,
    )

    # Pre-validate arguments.
    service_args = validate_and_load_arguments(module)

    # Choose logic based on the desired state.
    if service_args.state == "present":
        start_service(service_args, module)
    elif service_args.state == "absent":
        stop_service(service_args, module)
    elif service_args.state == "restarted":
        restart_service(service_args, module)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2019, Evgeniy Krysanov <evgeniy.krysanov@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: bitbucket_pipeline_variable
short_description: Manages Bitbucket pipeline variables
description:
  - Manages Bitbucket pipeline variables.
author:
  - Evgeniy Krysanov (@catcombo)
extends_documentation_fragment:
  - community.general._bitbucket
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  repository:
    description:
      - The repository name.
    type: str
    required: true
  workspace:
    description:
      - The repository owner.
      - B(Note:) O(ignore:username) used to be an alias of this option. Since community.general 6.0.0 it is an alias of O(user).
    type: str
    required: true
  name:
    description:
      - The pipeline variable name.
    type: str
    required: true
  value:
    description:
      - The pipeline variable value.
    type: str
  secured:
    description:
      - Whether to encrypt the variable value.
    type: bool
    default: false
  state:
    description:
      - Indicates desired state of the variable.
    type: str
    required: true
    choices: [absent, present]
notes:
  - Check mode is supported.
  - For secured values return parameter RV(ignore:changed) is always V(true).
"""

EXAMPLES = r"""
- name: Create or update pipeline variables from the list
  community.general.bitbucket_pipeline_variable:
    repository: 'bitbucket-repo'
    workspace: bitbucket_workspace
    name: '{{ item.name }}'
    value: '{{ item.value }}'
    secured: '{{ item.secured }}'
    state: present
  with_items:
    - {name: AWS_ACCESS_KEY, value: ABCD1234, secured: false}
    - {name: AWS_SECRET, value: qwe789poi123vbn0, secured: true}

- name: Remove pipeline variable
  community.general.bitbucket_pipeline_variable:
    repository: bitbucket-repo
    workspace: bitbucket_workspace
    name: AWS_ACCESS_KEY
    state: absent
"""

RETURN = r""" # """

from ansible.module_utils.basic import AnsibleModule, _load_params

from ansible_collections.community.general.plugins.module_utils._bitbucket import BitbucketHelper

error_messages = {
    "required_value": "`value` is required when the `state` is `present`",
}

BITBUCKET_API_ENDPOINTS = {
    "pipeline-variable-list": f"{BitbucketHelper.BITBUCKET_API_URL}/2.0/repositories/{{workspace}}/{{repo_slug}}/pipelines_config/variables/",
    "pipeline-variable-detail": (
        f"{BitbucketHelper.BITBUCKET_API_URL}/2.0/repositories/{{workspace}}/{{repo_slug}}/pipelines_config/variables/{{variable_uuid}}"
    ),
}


def get_existing_pipeline_variable(module, bitbucket):
    """
    Search for a pipeline variable

    :param module: instance of the :class:`AnsibleModule`
    :param bitbucket: instance of the :class:`BitbucketHelper`
    :return: existing variable or None if not found
    :rtype: dict or None

    Return example::

        {
            'name': 'AWS_ACCESS_OBKEY_ID',
            'value': 'x7HU80-a2',
            'type': 'pipeline_variable',
            'secured': False,
            'uuid': '{9ddb0507-439a-495a-99f3-5464f15128127}'
        }

    The `value` key in dict is absent in case of secured variable.
    """
    variables_base_url = BITBUCKET_API_ENDPOINTS["pipeline-variable-list"].format(
        workspace=module.params["workspace"],
        repo_slug=module.params["repository"],
    )
    # Look through the all response pages in search of variable we need
    page = 1
    while True:
        next_url = f"{variables_base_url}?page={page}"
        info, content = bitbucket.request(
            api_url=next_url,
            method="GET",
        )

        if info["status"] == 404:
            module.fail_json(msg="Invalid `repository` or `workspace`.")

        if info["status"] != 200:
            module.fail_json(msg=f"Failed to retrieve the list of pipeline variables: {info}")

        # We are at the end of list
        if "pagelen" in content and content["pagelen"] == 0:
            return None

        page += 1
        var = next((v for v in content["values"] if v["key"] == module.params["name"]), None)

        if var is not None:
            var["name"] = var.pop("key")
            return var


def create_pipeline_variable(module, bitbucket):
    info, content = bitbucket.request(
        api_url=BITBUCKET_API_ENDPOINTS["pipeline-variable-list"].format(
            workspace=module.params["workspace"],
            repo_slug=module.params["repository"],
        ),
        method="POST",
        data={
            "key": module.params["name"],
            "value": module.params["value"],
            "secured": module.params["secured"],
        },
    )

    if info["status"] != 201:
        module.fail_json(msg=f"Failed to create pipeline variable `{module.params['name']}`: {info}")


def update_pipeline_variable(module, bitbucket, variable_uuid):
    info, content = bitbucket.request(
        api_url=BITBUCKET_API_ENDPOINTS["pipeline-variable-detail"].format(
            workspace=module.params["workspace"],
            repo_slug=module.params["repository"],
            variable_uuid=variable_uuid,
        ),
        method="PUT",
        data={
            "value": module.params["value"],
            "secured": module.params["secured"],
        },
    )

    if info["status"] != 200:
        module.fail_json(msg=f"Failed to update pipeline variable `{module.params['name']}`: {info}")


def delete_pipeline_variable(module, bitbucket, variable_uuid):
    info, content = bitbucket.request(
        api_url=BITBUCKET_API_ENDPOINTS["pipeline-variable-detail"].format(
            workspace=module.params["workspace"],
            repo_slug=module.params["repository"],
            variable_uuid=variable_uuid,
        ),
        method="DELETE",
    )

    if info["status"] != 204:
        module.fail_json(msg=f"Failed to delete pipeline variable `{module.params['name']}`: {info}")


class BitBucketPipelineVariable(AnsibleModule):
    def __init__(self, *args, **kwargs):
        params = _load_params() or {}
        if params.get("secured"):
            kwargs["argument_spec"]["value"].update({"no_log": True})
        super().__init__(*args, **kwargs)


def main():
    argument_spec = BitbucketHelper.bitbucket_argument_spec()
    argument_spec.update(
        repository=dict(type="str", required=True),
        workspace=dict(type="str", required=True),
        name=dict(type="str", required=True),
        value=dict(type="str"),
        secured=dict(type="bool", default=False),
        state=dict(type="str", choices=["present", "absent"], required=True),
    )
    module = BitBucketPipelineVariable(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=BitbucketHelper.bitbucket_required_one_of(),
        required_together=BitbucketHelper.bitbucket_required_together(),
    )

    bitbucket = BitbucketHelper(module)

    value = module.params["value"]
    state = module.params["state"]
    secured = module.params["secured"]

    # Check parameters
    if (value is None) and (state == "present"):
        module.fail_json(msg=error_messages["required_value"])

    # Retrieve access token for authorized API requests
    bitbucket.fetch_access_token()

    # Retrieve existing pipeline variable (if any)
    existing_variable = get_existing_pipeline_variable(module, bitbucket)
    changed = False

    # Create new variable in case it doesn't exists
    if not existing_variable and (state == "present"):
        if not module.check_mode:
            create_pipeline_variable(module, bitbucket)
        changed = True

    # Update variable if it is secured or the old value does not match the new one
    elif existing_variable and (state == "present"):
        if (existing_variable["secured"] != secured) or (existing_variable.get("value") != value):
            if not module.check_mode:
                update_pipeline_variable(module, bitbucket, existing_variable["uuid"])
            changed = True

    # Delete variable
    elif existing_variable and (state == "absent"):
        if not module.check_mode:
            delete_pipeline_variable(module, bitbucket, existing_variable["uuid"])
        changed = True

    module.exit_json(changed=changed)


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the resource
#   module builder playbook.
#
# Do not edit this file manually.
#
# Changes to this file will be over written
#   by the resource module builder.
#
# Changes should be made in the model used to
#   generate this file or in the resource module
#   builder template.
#
#############################################
"""
The arg spec for the nxos_lldp_global module
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type


class Lldp_globalArgs(object):  # pylint: disable=R0903
    """The arg spec for the nxos_lldp_global module"""

    argument_spec = {
        "running_config": {"type": "str"},
        "config": {
            "options": {
                "holdtime": {"type": "int"},
                "port_id": {"choices": [0, 1], "type": "int"},
                "reinit": {"type": "int"},
                "timer": {"type": "int"},
                "tlv_select": {
                    "options": {
                        "dcbxp": {"type": "bool"},
                        "management_address": {
                            "options": {
                                "v4": {"type": "bool"},
                                "v6": {"type": "bool"},
                            },
                            "type": "dict",
                        },
                        "port": {
                            "options": {
                                "description": {"type": "bool"},
                                "vlan": {"type": "bool"},
                            },
                            "type": "dict",
                        },
                        "power_management": {"type": "bool"},
                        "system": {
                            "options": {
                                "capabilities": {"type": "bool"},
                                "description": {"type": "bool"},
                                "name": {"type": "bool"},
                            },
                            "type": "dict",
                        },
                    },
                    "type": "dict",
                },
            },
            "type": "dict",
        },
        "state": {
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "deleted",
                "gathered",
                "parsed",
                "rendered",
            ],
            "default": "merged",
            "type": "str",
        },
    }



#!/usr/bin/python
#
# Copyright (c) 2025 xuzhang3 (@xuzhang3), Fred-sun (@Fred-sun)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_keyvaultcertificate_info
version_added: "3.1.0"
short_description: Get Azure Key Vault certificate facts
description:
    - Get or list facts of Azure Key Vault certificate(deleted).

options:
    vault_uri:
        description:
            - Vault uri where the certificate stored in.
        required: True
        type: str
    name:
        description:
            - Certificate name. If not set, will list all certificates in vault_uri.
        type: str
    version:
        description:
            - The version of the certificate.
        type: str
    show_deleted_certificate:
        description:
            - Set to I(show_delete_certificate=true) to show deleted certificates. Set to I(show_deleted_certificate=false) to show not deleted certificates.
        type: bool
        default: false
    tags:
        description:
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - xuzhang3 (@xuzhang3)
    - Fred-sun (@Fred-sun)

'''

EXAMPLES = '''
- name: Get certificate facts
  azure_rm_keyvaultcertificate_info:
    vault_uri: "https://myVault.vault.azure.net"
    name: myCertificate

- name: Get specific versions of certificate
  azure_rm_keyvaultcertificate_info:
    vault_uri: "https://myVault.vault.azure.net"
    name: mySecret
    version: 2809225bcb674ff380f330471b3c3eb0

- name: Get deleted certificate
  azure_rm_keyvaultcertificate_info:
    vault_uri: "https://myVault.vault.azure.net"
    name: mySecret
    show_deleted_certificate: true

- name: List deleted certificate
  azure_rm_keyvaultcertificate_info:
    vault_uri: "https://myVault.vault.azure.net"
    show_deleted_certificate: true
'''

RETURN = '''
certificates:
    description:
        - The facts of certificates in Azure Key Vault.
    returned: always
    type: complex
    contains:
        cert_data:
            description:
                - CER contents of the X509 certificate.
            type: str
            returned: always
            sample: "MIID*****************0pRjXE"
        name:
            description:
                - The name of the certificate.
            type: str
            returned: always
            sample: testcert
        deleted_on:
            description:
                - The time when the certificate was deleted, in UTC.
            returned: always
            type: str
            sample: 2025-01-14T09
        recovery_id:
            description:
                - The url of the recovery object, used to identify and recover the deleted certificate.
            type: str
            returned: always
            sample: "https://vaultrfred01.vault.azure.net/deletedcertificates/cert02"
        scheduled_purge_date:
            description:
                - The time when the certificate is scheduled to be purged, in UTC.
            returned: always
            type: dict
            sample: 2025-02-14T09
        policy:
            description:
                - The management policy of the deleted certificate.
            returned: always
            type: complex
            contains:
                attributes:
                    description:
                        - Certificate attributes.
                    type: complex
                    returned: always
                    contains:
                        created:
                            description:
                                - Creation datetime.
                            returned: always
                            type: str
                            sample: "2025-01-14T09:41:20+00:00"
                        not_before:
                            description:
                                - Not before datetime.
                            type: str
                            sample: None
                        expires:
                            description:
                                - Expiration datetime.
                            type: str
                            sample: None
                        updated:
                            description:
                                - Update datetime.
                            returned: always
                            type: str
                            sample: "2025-01-15T09:41:20+00:00"
                        enabled:
                            description:
                                - Indicate whether the certificate is enabled.
                            returned: always
                            type: str
                            sample: true
                        recovery_level:
                            description:
                                - Reflects the deletion recovery level currently in effect for certificates in the current vault.
                                - If it contains 'Purgeable' the certificate can be permanently deleted by a privileged user,
                                - Otherwise, only the system can purge the certificate, at the end of the retention interval.
                            returned: always
                            type: str
                            sample: None
                        recoverable_days:
                            description:
                                - Reflects the deletion recovery days.
                            type: int
                            returned: always
                            sample: None
                issuer_name:
                    description:
                        - Name of the referenced issuer object or reserved names.
                    type: str
                    returned: always
                    sample: Self
                subject:
                    description:
                        - The subject name of the certificate.
                        - Should be a valid X509 distinguished name.
                        - Either subject or one of the subject alternative name parameters are required for creating a certificate.
                        - This will be ignored when importing a certificate; the subject will be parsed from the imported certificate.
                    type: str
                    returned: always
                    sample: CN=anhui.com
                san_emails:
                    description:
                        - Subject alternative emails of the X509 object.
                        - Either subject or one of the subject alternative name parameters are required for creating a certificate.
                    type: str
                    returned: always
                    sample: None
                san_dns_names:
                    description:
                        - Subject alternative DNS names of the X509 object.
                        - Either subject or one of the subject alternative name parameters are required for creating a certificate.
                    type: str
                    returned: always
                    sample: None
                san_user_principal_names:
                    description:
                        - Subject alternative user principal names of the X509 object.
                        - Either subject or one of the subject alternative name parameters are required for creating a certificate.
                    type: str
                    returned: always
                    sample: None
                exportable:
                    description:
                        - Indicates if the private key can be exported.
                    type: bool
                    returned: always
                    sample: true
                key_type:
                    description:
                        - The type of key pair to be used for the certificate.
                    type: str
                    returned: always
                    sample: RSA
                key_size:
                    description:
                        - The key size in bits.
                    type: int
                    returned: always
                    sample: 2048
                reuse_key:
                    description:
                        - Indicates if the same key pair will be used on certificate renewal.
                    type: bool
                    returned: always
                    sample: false
                key_curve_name:
                    description:
                        - Elliptic curve name. For valid values, see KeyCurveName.
                    type: str
                    returned: always
                    sample: None
                enhanced_key_usage:
                    description:
                        - The extended ways the key of the certificate can be used.
                    type: list
                    returned: always
                    sample: ['1.3.6.1.5.5.7.3.1', '1.3.6.1.5.5.7.3.2']
                key_usage:
                    description:
                        - List of key usages.
                    type: list
                    returned: always
                    sample: ["digitalSignature", "keyEncipherment"]
                content_type:
                    description:
                        - If not specified, the media type (MIME type) of the secret backing the certificate.
                    type: str
                    returned: always
                    sample: application/x-pkcs12
                validity_in_months:
                    description:
                        - The duration that the certificate is valid in months.
                    type: int
                    returned: always
                    sample: 12
                lifetime_actions:
                    description:
                        - Actions that will be performed by Key Vault over the lifetime of a certificate.
                    type: list
                    returned: always
                    sample: [{'action': 'AutoRenew', 'days_before_expiry': None, 'lifetime_percentage': 80}]
                certificate_type:
                    description:
                        - Type of certificate to be requested from the issuer provider.
                    type: str
                    returned: str
                    sample: None
                certificate_transparency:
                    description:
                        - Indicates if the certificates generated under this policy should be published to certificate transparency logs.
                    type: bool
                    returned: always
                    sample: None
        properties:
            description:
                - The certificate's properties.
            type: complex
            returned: always
            contains:
                id:
                    description:
                        - Id of the certificate. If specified all other 'Id' arguments should be omitted.
                    type: str
                    returned: always
                    sample: "https://vaultrfred01.vault.azure.net/certificates/cert02/62409e6304c642f193209729b8360d2c"
                vault_id:
                    description:
                        - ID of the Key Vault.
                    type: str
                    returned: always
                    sample: "https://vaultrfred01.vault.azure.net"
                x509_thumbprint:
                    description:
                        - The X509 Thumbprint of the Key Vault Certificate represented as a hexadecimal string.
                    type: str
                    returned: always
                    sample: 1blAnHN9ddng0qh1pYoUDY2lp1E=
                tags:
                    description:
                        - List of the certificate tags.
                    type: dict
                    returned: always
                    sample: {'key': 'value'}
                attributes:
                    description:
                        - Certificate attributes.
                    type: complex
                    returned: always
                    contains:
                        created:
                            description:
                                - Creation datetime.
                            returned: always
                            type: str
                            sample: "2025-01-14T09"
                        not_before:
                            description:
                                - Not before datetime.
                            type: str
                            sample: "2025-02-14T09"
                        expires:
                            description:
                                - Expiration datetime.
                            type: str
                            sample: "2025-03-14T09"
                        updated:
                            description:
                                - Update datetime.
                            returned: always
                            type: str
                            sample: "2025-01-15T09"
                        enabled:
                            description:
                                - Indicate whether the certificate is enabled.
                            returned: always
                            type: str
                            sample: true
                        recovery_level:
                            description:
                                - Reflects the deletion recovery level currently in effect for certificates in the current vault.
                                - If it contains 'Purgeable' the certificate can be permanently deleted by a privileged user,
                                - Otherwise, only the system can purge the certificate, at the end of the retention interval.
                            returned: always
                            type: str
                            sample: Recoverable+Purgeable
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase

try:
    from azure.keyvault.certificates import CertificateClient
    from azure.core.exceptions import ResourceNotFoundError
    import base64
except ImportError:
    # This is handled in azure_rm_common
    pass


def certificatebundle_to_dict(certificate):
    response = dict(policy=dict(), properties=dict(), cert_data=None)
    if certificate.cer is not None:
        response['cert_data'] = base64.b64encode(certificate.cer).decode('utf-8')
    response['name'] = certificate.name
    if certificate.policy is not None:
        response['policy']['issuer_name'] = certificate.policy._issuer_name
        response['policy']['subject'] = certificate.policy._subject
        response['policy']['exportable'] = certificate.policy._exportable
        response['policy']['key_type'] = certificate.policy._key_type
        response['policy']['reuse_key'] = certificate.policy._reuse_key
        response['policy']['key_size'] = certificate.policy._key_size
        response['policy']['key_curve_name'] = certificate.policy._key_curve_name
        response['policy']['enhanced_key_usage'] = certificate.policy._enhanced_key_usage
        response['policy']['key_usage'] = certificate.policy._key_usage
        response['policy']['content_type'] = certificate.policy._content_type
        response['policy']['validity_in_months'] = certificate.policy._validity_in_months
        response['policy']['certificate_type'] = certificate.policy._certificate_type
        response['policy']['certificate_transparency'] = certificate.policy._certificate_transparency
        response['policy']['san_emails'] = certificate.policy._san_emails
        response['policy']['san_dns_names'] = certificate.policy._san_dns_names
        response['policy']['san_user_principal_names'] = certificate.policy._san_user_principal_names
        response['policy']['attributes'] = dict()
        if certificate.policy._attributes is not None:
            response['policy']['attributes']['enabled'] = certificate.policy._attributes.enabled
            response['policy']['attributes']['not_before'] = certificate.policy._attributes.not_before
            response['policy']['attributes']['expires'] = certificate.policy._attributes.expires
            response['policy']['attributes']['created'] = certificate.policy._attributes.created
            response['policy']['attributes']['updated'] = certificate.policy._attributes.updated
            response['policy']['attributes']['recoverable_days'] = certificate.policy._attributes.recoverable_days
            response['policy']['attributes']['recovery_level'] = certificate.policy._attributes.recovery_level
        else:
            response['policy']['attributes'] = None
        if certificate.policy._lifetime_actions is not None:
            response['policy']['lifetime_actions'] = []
            for item in certificate.policy._lifetime_actions:
                response['policy']['lifetime_actions'].append(dict(action=item.action,
                                                                   lifetime_percentage=item.lifetime_percentage,
                                                                   days_before_expiry=item.days_before_expiry))
        else:
            response['policy']['lifetime_actions'] = None
    else:
        response['policy'] = None

    if certificate.properties is not None:
        response['properties']['attributes'] = dict(enabled=certificate.properties._attributes.enabled,
                                                    not_before=certificate.properties._attributes.not_before,
                                                    expires=certificate.properties._attributes.expires,
                                                    created=certificate.properties._attributes.created,
                                                    updated=certificate.properties._attributes.updated,
                                                    recovery_level=certificate.properties._attributes.recovery_level)
        response['properties']['id'] = certificate.properties._id
        response['properties']['vault_id'] = certificate.properties._vault_id.vault_url if certificate.properties._vault_id is not None else None
        response['properties']['x509_thumbprint'] = base64.b64encode(certificate.properties._x509_thumbprint).decode('utf-8')
        response['properties']['tags'] = certificate.properties._tags
    else:
        response['properties'] = None

    return response


def deleted_certificatebundle_to_dict(certificate):
    response = dict(policy=dict(), properties=dict(), cert_data=None)
    if certificate.cer is not None:
        response['cert_data'] = str(certificate.cer)
    response['name'] = certificate.name
    response['recovery_id'] = certificate._recovery_id
    response['scheduled_purge_date'] = certificate._scheduled_purge_date
    response['deleted_on'] = certificate._deleted_on
    if certificate.policy is not None:
        response['policy']['issuer_name'] = certificate.policy._issuer_name
        response['policy']['subject'] = certificate.policy._subject
        response['policy']['exportable'] = certificate.policy._exportable
        response['policy']['key_type'] = certificate.policy._key_type
        response['policy']['key_size'] = certificate.policy._key_size
        response['policy']['reuse_key'] = certificate.policy._reuse_key
        response['policy']['key_curve_name'] = certificate.policy._key_curve_name
        response['policy']['enhanced_key_usage'] = certificate.policy._enhanced_key_usage
        response['policy']['key_usage'] = certificate.policy._key_usage
        response['policy']['content_type'] = certificate.policy._content_type
        response['policy']['validity_in_months'] = certificate.policy._validity_in_months
        response['policy']['certificate_type'] = certificate.policy._certificate_type
        response['policy']['certificate_transparency'] = certificate.policy._certificate_transparency
        response['policy']['san_emails'] = certificate.policy._san_emails
        response['policy']['san_dns_names'] = certificate.policy._san_dns_names
        response['policy']['san_user_principal_names'] = certificate.policy._san_user_principal_names
        response['policy']['attributes'] = dict()
        if certificate.policy._attributes is not None:
            response['policy']['attributes']['enabled'] = certificate.policy._attributes.enabled
            response['policy']['attributes']['not_before'] = certificate.policy._attributes.not_before
            response['policy']['attributes']['expires'] = certificate.policy._attributes.expires
            response['policy']['attributes']['created'] = certificate.policy._attributes.created
            response['policy']['attributes']['updated'] = certificate.policy._attributes.updated
            response['policy']['attributes']['recoverable_days'] = certificate.policy._attributes.recoverable_days
            response['policy']['attributes']['recovery_level'] = certificate.policy._attributes.recovery_level
        else:
            response['policy']['attributes'] = None
        if certificate.policy._lifetime_actions is not None:
            response['policy']['lifetime_actions'] = []
            for item in certificate.policy._lifetime_actions:
                response['policy']['lifetime_actions'].append(dict(action=item.action,
                                                                   lifetime_percentage=item.lifetime_percentage,
                                                                   days_before_expiry=item.days_before_expiry))
        else:
            response['policy']['lifetime_actions'] = None
    else:
        response['policy'] = None

    if certificate.properties is not None:
        response['properties']['attributes'] = dict(enabled=certificate.properties._attributes.enabled,
                                                    not_before=certificate.properties._attributes.not_before,
                                                    expires=certificate.properties._attributes.expires,
                                                    created=certificate.properties._attributes.created,
                                                    updated=certificate.properties._attributes.updated,
                                                    recovery_level=certificate.properties._attributes.recovery_level)
        response['properties']['id'] = certificate.properties._id
        response['properties']['vault_id'] = certificate.properties._vault_id.vault_url if certificate.properties._vault_id is not None else None
        response['properties']['x509_thumbprint'] = base64.b64encode(certificate.properties._x509_thumbprint).decode('utf-8')
        response['properties']['tags'] = certificate.properties._tags
    else:
        response['properties'] = None
    return response


class AzureRMKeyVaultCertificateInfo(AzureRMModuleBase):
    def __init__(self):
        self.module_arg_spec = dict(version=dict(type='str'),
                                    name=dict(type='str'),
                                    vault_uri=dict(type='str', required=True),
                                    show_deleted_certificate=dict(type='bool', default=False),
                                    tags=dict(type='list', elements='str'))

        self.vault_uri = None
        self.name = None
        self.version = None
        self.show_deleted_certificate = False
        self.tags = None

        self.results = dict(changed=False)
        self._client = None

        super(AzureRMKeyVaultCertificateInfo,
              self).__init__(derived_arg_spec=self.module_arg_spec,
                             supports_check_mode=True,
                             supports_tags=False,
                             facts_module=True)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for certificate in list(self.module_arg_spec.keys()):
            if hasattr(self, certificate):
                setattr(self, certificate, kwargs[certificate])

        self._client = self.get_keyvault_client()

        if self.name:
            if self.show_deleted_certificate:
                self.results['certificates'] = self.get_deleted_certificate()
            else:
                if self.version is not None:
                    self.results['certificates'] = self.get_certificate_version()
                else:
                    self.results['certificates'] = self.get_certificate()
        else:
            if self.show_deleted_certificate:
                self.results['certificates'] = self.list_deleted_certificates()

        return self.results

    def get_keyvault_client(self):

        return CertificateClient(vault_url=self.vault_uri, credential=self.azure_auth.azure_credential_track2)

    def get_certificate(self):
        '''
        Gets the certificate fact of the specified certificate in key vault.

        :return: deserialized certificate state dictionary
        '''
        self.log("Get the certificate {0}".format(self.name))

        results = []
        try:
            response = self._client.get_certificate(certificate_name=self.name)

            if response:
                response = certificatebundle_to_dict(response)
                if self.has_tags(response['properties']['tags'], self.tags):
                    self.log("Response : {0}".format(response))
                    results.append(response)

        except ResourceNotFoundError as ec:
            self.log("Did not find the key vault certificate {0}: {1}".format(
                self.name, str(ec)))
        except Exception as ec2:
            self.fail("Find the key vault certificate got exception, exception as {0}".format(str(ec2)))
        return results

    def get_deleted_certificate(self):
        '''
        Gets the deleted certificate facts in key vault.

        :return: deserialized certificate state dictionary
        '''
        self.log("Get the certificate {0}".format(self.name))

        results = []
        try:
            response = self._client.get_deleted_certificate(certificate_name=self.name)
            if response:
                response = deleted_certificatebundle_to_dict(response)
                if self.has_tags(response['properties'].get('tags'), self.tags):
                    self.log("Response : {0}".format(response))
                    results.append(response)

        except ResourceNotFoundError as ec:
            self.log("Did not find the key vault certificate {0}: {1}".format(self.name, str(ec)))
        except Exception as ec2:
            self.fail("Find the key vault certificate got exception, exception as {0}".format(str(ec2)))
        return results

    def get_certificate_version(self):
        '''
        Lists certificates versions.

        :return: deserialized versions of certificate, includes certificate identifier, attributes and tags
        '''
        self.log("Get the certificate versions {0}".format(self.name))

        try:
            response = self._client.get_certificate_version(certificate_name=self.name, version=self.version)
            self.log("Response : {0}".format(response))

            if response:
                res = certificatebundle_to_dict(response)
                if self.has_tags(res['properties'].get('tags'), self.tags):
                    return res
        except Exception as e:
            self.fail("Did not find certificate versions {0} : {1}.".format(
                self.name, str(e)))

    def list_deleted_certificates(self):
        '''
        Lists deleted certificates in specific key vault.

        :return: deserialized certificates, includes certificate identifier, attributes and tags.
        '''
        self.log("Get the key vaults in current subscription")

        results = []
        try:
            response = self._client.list_deleted_certificates()
            self.log("Response : {0}".format(response))

            if response:
                for item in response:
                    item = deleted_certificatebundle_to_dict(item)
                    if self.has_tags(item['properties'].get('tags'), self.tags):
                        results.append(item)
        except Exception as e:
            self.fail("Did not find certificate in current key vault {0}.".format(str(e)))
        return results


def main():
    """Main execution"""
    AzureRMKeyVaultCertificateInfo()


if __name__ == '__main__':
    main()



# Copyright (c) 2020, Doug Stanley <doug+ansible@technologixllc.com>
# Copyright (c) 2021, Andrew Pantuso (@ajpantuso) <ajpantuso@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import os
import re
import typing as t
from contextlib import contextmanager
from struct import Struct

# Protocol References
# -------------------
# https://datatracker.ietf.org/doc/html/rfc4251
# https://datatracker.ietf.org/doc/html/rfc4253
# https://datatracker.ietf.org/doc/html/rfc5656
# https://datatracker.ietf.org/doc/html/rfc8032
#
# Inspired by:
# ------------
# https://github.com/pyca/cryptography/blob/main/src/cryptography/hazmat/primitives/serialization/ssh.py
# https://github.com/paramiko/paramiko/blob/master/paramiko/message.py

# 0 (False) or 1 (True) encoded as a single byte
_BOOLEAN = Struct(b"?")
# Unsigned 8-bit integer in network-byte-order
_UBYTE = Struct(b"!B")
_UBYTE_MAX = 0xFF
# Unsigned 32-bit integer in network-byte-order
_UINT32 = Struct(b"!I")
# Unsigned 32-bit little endian integer
_UINT32_LE = Struct(b"<I")
_UINT32_MAX = 0xFFFFFFFF
# Unsigned 64-bit integer in network-byte-order
_UINT64 = Struct(b"!Q")
_UINT64_MAX = 0xFFFFFFFFFFFFFFFF


_T = t.TypeVar("_T")


def any_in(sequence: t.Iterable[_T], *elements: _T) -> bool:
    return any(e in sequence for e in elements)


def file_mode(path: str | os.PathLike) -> int:
    if not os.path.exists(path):
        return 0o000
    return os.stat(path).st_mode & 0o777


def parse_openssh_version(version_string: str) -> str | None:
    """Parse the version output of ssh -V and return version numbers that can be compared"""

    parsed_result = re.match(
        r"^.*openssh_(?P<version>[0-9.]+)(p?[0-9]+)[^0-9]*.*$", version_string.lower()
    )
    if parsed_result is not None:
        version = parsed_result.group("version").strip()
    else:
        version = None

    return version


@contextmanager
def secure_open(*, path: str | os.PathLike, mode: int) -> t.Iterator[int]:
    fd = os.open(path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, mode)
    try:
        yield fd
    finally:
        os.close(fd)


def secure_write(*, path: str | os.PathLike, mode: int, content: bytes) -> None:
    with secure_open(path=path, mode=mode) as fd:
        os.write(fd, content)


# See https://datatracker.ietf.org/doc/html/rfc4251#section-5 for SSH data types
class OpensshParser:
    """Parser for OpenSSH encoded objects"""

    BOOLEAN_OFFSET = 1
    UINT32_OFFSET = 4
    UINT64_OFFSET = 8

    def __init__(self, *, data: bytes | bytearray) -> None:
        if not isinstance(data, (bytes, bytearray)):
            raise TypeError(f"Data must be bytes-like not {type(data)}")

        self._data = memoryview(data)
        self._pos = 0

    def boolean(self) -> bool:
        next_pos = self._check_position(self.BOOLEAN_OFFSET)

        value: bool = _BOOLEAN.unpack(self._data[self._pos : next_pos])[0]
        self._pos = next_pos
        return value

    def uint32(self) -> int:
        next_pos = self._check_position(self.UINT32_OFFSET)

        value: int = _UINT32.unpack(self._data[self._pos : next_pos])[0]
        self._pos = next_pos
        return value

    def uint64(self) -> int:
        next_pos = self._check_position(self.UINT64_OFFSET)

        value: int = _UINT64.unpack(self._data[self._pos : next_pos])[0]
        self._pos = next_pos
        return value

    def string(self) -> bytes:
        length = self.uint32()

        next_pos = self._check_position(length)

        value = self._data[self._pos : next_pos]
        self._pos = next_pos
        # Cast to bytes is required as a memoryview slice is itself a memoryview
        return bytes(value)

    def mpint(self) -> int:
        return self._big_int(self.string(), "big", signed=True)

    def name_list(self) -> list[str]:
        raw_string = self.string()
        return raw_string.decode("ASCII").split(",")

    # Convenience function, but not an official data type from SSH
    def string_list(self) -> list[bytes]:
        result = []
        raw_string = self.string()

        if raw_string:
            parser = OpensshParser(data=raw_string)
            while parser.remaining_bytes():
                result.append(parser.string())

        return result

    # Convenience function, but not an official data type from SSH
    def option_list(self) -> list[tuple[bytes, bytes]]:
        result = []
        raw_string = self.string()

        if raw_string:
            parser = OpensshParser(data=raw_string)

            while parser.remaining_bytes():
                name = parser.string()
                data = parser.string()
                if data:
                    # data is doubly-encoded
                    data = OpensshParser(data=data).string()
                result.append((name, data))

        return result

    def seek(self, offset: int) -> int:
        self._pos = self._check_position(offset)

        return self._pos

    def remaining_bytes(self) -> int:
        return len(self._data) - self._pos

    def _check_position(self, offset: int) -> int:
        if self._pos + offset > len(self._data):
            raise ValueError(f"Insufficient data remaining at position: {self._pos}")
        if self._pos + offset < 0:
            raise ValueError("Position cannot be less than zero.")
        return self._pos + offset

    @classmethod
    def signature_data(cls, *, signature_string: bytes) -> dict[str, bytes | int]:
        signature_data: dict[str, bytes | int] = {}

        parser = cls(data=signature_string)
        signature_type = parser.string()
        signature_blob = parser.string()

        blob_parser = cls(data=signature_blob)
        if signature_type in (b"ssh-rsa", b"rsa-sha2-256", b"rsa-sha2-512"):
            # https://datatracker.ietf.org/doc/html/rfc4253#section-6.6
            # https://datatracker.ietf.org/doc/html/rfc8332#section-3
            signature_data["s"] = cls._big_int(signature_blob, "big")
        elif signature_type == b"ssh-dss":
            # https://datatracker.ietf.org/doc/html/rfc4253#section-6.6
            signature_data["r"] = cls._big_int(signature_blob[:20], "big")
            signature_data["s"] = cls._big_int(signature_blob[20:], "big")
        elif signature_type in (
            b"ecdsa-sha2-nistp256",
            b"ecdsa-sha2-nistp384",
            b"ecdsa-sha2-nistp521",
        ):
            # https://datatracker.ietf.org/doc/html/rfc5656#section-3.1.2
            signature_data["r"] = blob_parser.mpint()
            signature_data["s"] = blob_parser.mpint()
        elif signature_type == b"ssh-ed25519":
            # https://datatracker.ietf.org/doc/html/rfc8032#section-5.1.2
            signature_data["R"] = cls._big_int(signature_blob[:32], "little")
            signature_data["S"] = cls._big_int(signature_blob[32:], "little")
        else:
            raise ValueError(f"{signature_type!r} is not a valid signature type")

        signature_data["signature_type"] = signature_type

        return signature_data

    @classmethod
    def _big_int(
        cls,
        raw_string: bytes,
        byte_order: t.Literal["big", "little"],
        signed: bool = False,
    ) -> int:
        if byte_order not in ("big", "little"):
            raise ValueError(
                f"Byte_order must be one of (big, little) not {byte_order}"
            )

        return int.from_bytes(raw_string, byte_order, signed=signed)


class _OpensshWriter:
    """Writes SSH encoded values to a bytes-like buffer

    .. warning::
        This class is a private API and must not be exported outside of the openssh module_utils.
        It is not to be used to construct Openssh objects, but rather as a utility to assist
        in validating parsed material.
    """

    def __init__(self, *, buffer: bytearray | None = None):
        if buffer is not None:
            if not isinstance(buffer, bytearray):
                raise TypeError(f"Buffer must be a bytearray, not {type(buffer)}")
        else:
            buffer = bytearray()

        self._buff: bytearray = buffer

    def boolean(self, value: bool) -> t.Self:
        if not isinstance(value, bool):
            raise TypeError(f"Value must be of type bool not {type(value)}")

        self._buff.extend(_BOOLEAN.pack(value))

        return self

    def uint32(self, value: int) -> t.Self:
        if not isinstance(value, int):
            raise TypeError(f"Value must be of type int not {type(value)}")
        if value < 0 or value > _UINT32_MAX:
            raise ValueError(
                f"Value must be a positive integer less than {_UINT32_MAX}"
            )

        self._buff.extend(_UINT32.pack(value))

        return self

    def uint64(self, value: int) -> t.Self:
        if not isinstance(value, int):
            raise TypeError(f"Value must be of type int not {type(value)}")
        if value < 0 or value > _UINT64_MAX:
            raise ValueError(
                f"Value must be a positive integer less than {_UINT64_MAX}"
            )

        self._buff.extend(_UINT64.pack(value))

        return self

    def string(self, value: bytes | bytearray) -> t.Self:
        if not isinstance(value, (bytes, bytearray)):
            raise TypeError(f"Value must be bytes-like not {type(value)}")
        self.uint32(len(value))
        self._buff.extend(value)

        return self

    def mpint(self, value: int) -> t.Self:
        if not isinstance(value, int):
            raise TypeError(f"Value must be of type int not {type(value)}")

        self.string(self._int_to_mpint(value))

        return self

    def name_list(self, value: list[str]) -> t.Self:
        if not isinstance(value, list):
            raise TypeError(f"Value must be a list of byte strings not {type(value)}")

        try:
            self.string(",".join(value).encode("ASCII"))
        except UnicodeEncodeError as e:
            raise ValueError(
                f"Name-list's must consist of US-ASCII characters: {e}"
            ) from e

        return self

    def string_list(self, value: list[bytes]) -> t.Self:
        if not isinstance(value, list):
            raise TypeError(f"Value must be a list of byte string not {type(value)}")

        writer = _OpensshWriter()
        for s in value:
            writer.string(s)

        self.string(writer.bytes())

        return self

    def option_list(self, value: list[tuple[bytes, bytes]]) -> t.Self:
        if not isinstance(value, list) or (value and not isinstance(value[0], tuple)):
            raise TypeError("Value must be a list of tuples")

        writer = _OpensshWriter()
        for name, data in value:
            writer.string(name)
            # SSH option data is encoded twice though this behavior is not documented
            writer.string(_OpensshWriter().string(data).bytes() if data else b"")

        self.string(writer.bytes())

        return self

    @staticmethod
    def _int_to_mpint(num: int) -> bytes:
        byte_length = (num.bit_length() + 7) // 8
        try:
            return num.to_bytes(byte_length, "big", signed=True)
        # Handles values which require \x00 or \xFF to pad sign-bit
        except OverflowError:
            return num.to_bytes(byte_length + 1, "big", signed=True)

    def bytes(self) -> bytes:
        return bytes(self._buff)


__all__ = (
    "OpensshParser",
    "any_in",
    "file_mode",
    "parse_openssh_version",
    "secure_open",
    "secure_write",
)



#
# (c) 2016 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import absolute_import, division, print_function

__metaclass__ = type

import json
import re

from ansible.errors import AnsibleConnectionFailure
from ansible.module_utils._text import to_text, to_bytes
from ansible.plugins.terminal import TerminalBase
from ansible.utils.display import Display

display = Display()


class TerminalModule(TerminalBase):
    # https://docs.ansible.com/ansible/latest/collections/ansible/netcommon/network_cli_connection.html

    terminal_stdout_re = [
        re.compile(rb"[\r\n]?[\w\+\-\.:\/\[\]]+(?:\([^\)]+\)){0,3}(?:[>#]) ?$")
    ]

    terminal_stderr_re = [
        re.compile(rb"% ?Error"),
        re.compile(rb"^% \w+", re.M),
        re.compile(rb"% ?Bad secret"),
        re.compile(rb"[\r\n%] Bad passwords"),
        re.compile(rb"invalid input", re.I),
        re.compile(rb"(?:incomplete|ambiguous) command", re.I),
        re.compile(rb"connection timed out", re.I),
        re.compile(rb"[^\r\n]+ not found"),
        re.compile(rb"'[^']' +returned error code: ?\d+"),
        re.compile(rb"Bad mask", re.I),
        re.compile(rb"% ?(\S+) ?overlaps with ?(\S+)", re.I),
        re.compile(rb"[%\S] ?Error: ?[\s]+", re.I),
        re.compile(rb"[%\S] ?Informational: ?[\s]+", re.I),
        re.compile(rb"Command authorization failed"),
    ]

    def on_open_shell(self):
        try:
            self._exec_cli_command(b"terminal datadump")
        except AnsibleConnectionFailure as e:
            raise AnsibleConnectionFailure("unable to set terminal parameters") from e
        try:
            self._exec_cli_command(b"terminal width 0")
        except AnsibleConnectionFailure:
            display.display(
                "WARNING: Unable to set terminal width, command responses may be truncated"
            )

        try:
            self._exec_cli_command(b"terminal no prompt")
        except AnsibleConnectionFailure:
            display.display(
                "WARNING: Unable disable prompt, command responses may fail"
            )

    def on_become(self, passwd=None):
        cmd = {"command": "enable"}
        if passwd:
            # Note: python-3.5 cannot combine u"" and r"" together.  Thus make
            # an r string and use to_text to ensure it's text on both py2 and py3.
            cmd["prompt"] = to_text(
                r"[\r\n]?(?:.*)?[Pp]assword: ?$", errors="surrogate_or_strict"
            )
            cmd["answer"] = passwd
            cmd["prompt_retry_check"] = True
        try:
            self._exec_cli_command(
                to_bytes(json.dumps(cmd), errors="surrogate_or_strict")
            )
            prompt = self._get_prompt()
            if prompt is None or not prompt.endswith(b"#"):
                raise AnsibleConnectionFailure(
                    "failed to elevate privilege to enable mode still at prompt [%s]"
                    % prompt
                )
        except AnsibleConnectionFailure as e:
            prompt = self._get_prompt()
            raise AnsibleConnectionFailure(
                "unable to elevate privilege to enable mode, at prompt [%s] with error: %s"
                % (prompt, e.message)
            ) from e

    def on_unbecome(self):
        prompt = self._get_prompt()
        if prompt is None:
            # if prompt is None most likely the terminal is hung up at a prompt
            return

        if b"(config" in prompt:
            self._exec_cli_command(b"end")
            self._exec_cli_command(b"disable")

        elif prompt.endswith(b"#"):
            self._exec_cli_command(b"disable")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2023, Tim Cragg (@timcragg) <tcragg@cisco.com>
# Copyright: (c) 2023, Akini Ross (@akinross) <akinross@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_system_banner
short_description: Manages Alias and Banners (aaa:PreLoginBanner)
description:
- Manages Alias and Banners on Cisco ACI fabrics.
options:
  description:
    description:
    - The description of the AAA login banner.
    type: str
  gui_alias:
    description:
    - The system GUI alias.
    type: str
  controller_banner:
    description:
    - The contents of the CLI informational banner to be displayed before user login authentication.
    - The CLI banner is a text based string printed as-is to the console.
    type: str
  switch_banner:
    description:
    - The switch banner message.
    type: str
  application_banner:
    description:
    - The application banner message.
    type: str
  severity:
    description:
    - The application banner severity.
    type: str
    choices: [ critical, info, major, minor, warning ]
  gui_banner:
    description:
    - The contents of the GUI informational banner to be displayed before user login authentication.
    - When I(gui_banner) starts with I(http://) or I(https://) the banner will be of URL type.
    - Note that the URL site owner must allow the site to be placed in an iFrame to display the informational banner.
    type: str
  state:
    description:
    - Use C(present) for updating.
    - Use C(query) for listing an object.
    type: str
    choices: [ present, query ]
    default: present
  name_alias:
    description:
    - The alias for the current object. This relates to the nameAlias field in ACI.
    type: str
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation
- cisco.aci.owner

seealso:
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(aaa:PreLoginBanner).
  link: https://developer.cisco.com/docs/apic-mim-ref/

author:
- Tim Cragg (@timcragg)
- Akini Ross (@akinross)
"""

EXAMPLES = r"""
- name: Configure banner
  cisco.aci.aci_system_banner:
    host: apic
    username: admin
    password: SomeSecretPassword
    gui_alias: "Test GUI Alias"
    controller_banner: "Test Controller Banner"
    application_banner: "Test Application Banner"
    severity: critical
    switch_banner: "Test Switch Banner"
    gui_banner: "Test GUI Banner"
    state: present
  delegate_to: localhost

- name: Configure banner with a url
  cisco.aci.aci_system_banner:
    host: apic
    username: admin
    password: SomeSecretPassword
    gui_alias: "Test GUI Alias"
    controller_banner: "Test Controller Banner Message"
    application_banner: "Test Application Banner"
    severity: critical
    switch_banner: "Test Switch Banner Message"
    gui_banner: https://www.cisco.com
    state: present
  delegate_to: localhost

- name: Query banner
  cisco.aci.aci_system_banner:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""


import re
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec, aci_owner_spec


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(aci_owner_spec())
    argument_spec.update(
        description=dict(type="str"),
        gui_alias=dict(type="str"),
        controller_banner=dict(type="str"),
        switch_banner=dict(type="str"),
        application_banner=dict(type="str"),
        severity=dict(type="str", choices=["critical", "info", "major", "minor", "warning"]),
        gui_banner=dict(type="str"),
        name_alias=dict(type="str"),
        state=dict(type="str", default="present", choices=["present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_by={"severity": "application_banner"},
    )
    aci = ACIModule(module)

    description = module.params.get("description")
    gui_alias = module.params.get("gui_alias")
    controller_banner = module.params.get("controller_banner")
    switch_banner = module.params.get("switch_banner")
    application_banner = module.params.get("application_banner")
    severity = module.params.get("severity")
    gui_banner = module.params.get("gui_banner")
    state = module.params.get("state")
    name_alias = module.params.get("name_alias")

    aci.construct_url(
        root_class=dict(
            aci_class="aaaPreLoginBanner",
            aci_rn="userext/preloginbanner",
        ),
    )
    aci.get_existing()

    if state == "present":
        regex_url = "^https?:\\/\\/(?:www\\.)?[-a-zA-Z0-9@:%._\\+~#=]{1,256}\\.[a-zA-Z0-9()]{1,6}\\b(?:[-a-zA-Z0-9()@:%_\\+.~#?&\\/=]*)$"

        if gui_banner is not None and re.fullmatch(regex_url, gui_banner):
            is_gui_message_text = "no"
        else:
            is_gui_message_text = "yes"

        aci.payload(
            aci_class="aaaPreLoginBanner",
            class_config=dict(
                descr=description,
                bannerMessage=application_banner,
                bannerMessageSeverity=severity,
                guiMessage=gui_banner,
                guiTextMessage=gui_alias,
                isGuiMessageText=is_gui_message_text,
                message=controller_banner,
                showBannerMessage="yes" if application_banner else "no",
                switchMessage=switch_banner,
                nameAlias=name_alias,
            ),
        )

        aci.get_diff(aci_class="aaaPreLoginBanner")

        aci.post_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



# Copyright (c) 2019, Ximon Eighteen <ximon.eighteen@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: docker_machine
author: Ximon Eighteen (@ximon18)
short_description: Docker Machine inventory source
requirements:
  - L(Docker Machine,https://docs.docker.com/machine/)
extends_documentation_fragment:
  - ansible.builtin.constructed
  - community.library_inventory_filtering_v1.inventory_filter
description:
  - Get inventory hosts from Docker Machine.
  - Uses a YAML configuration file that ends with V(docker_machine.(yml|yaml\)).
  - The plugin sets standard host variables C(ansible_host), C(ansible_port), C(ansible_user) and C(ansible_ssh_private_key).
  - The plugin stores the Docker Machine 'env' output variables in C(dm_) prefixed host variables.
notes:
  - The configuration file must be a YAML file whose filename ends with V(docker_machine.yml) or V(docker_machine.yaml). Other
    filenames will not be accepted.
options:
  plugin:
    description: Token that ensures this is a source file for the C(docker_machine) plugin.
    required: true
    choices: ['docker_machine', 'community.docker.docker_machine']
  daemon_env:
    description:
      - Whether docker daemon connection environment variables should be fetched, and how to behave if they cannot be fetched.
      - With V(require) and V(require-silently), fetch them and skip any host for which they cannot be fetched. A warning
        will be issued for any skipped host if the choice is V(require).
      - With V(optional) and V(optional-silently), fetch them and not skip hosts for which they cannot be fetched. A warning
        will be issued for hosts where they cannot be fetched if the choice is V(optional).
      - With V(skip), do not attempt to fetch the docker daemon connection environment variables.
      - If fetched successfully, the variables will be prefixed with C(dm_) and stored as host variables.
    type: str
    choices:
      - require
      - require-silently
      - optional
      - optional-silently
      - skip
    default: require
  running_required:
    description:
      - When V(true), hosts which Docker Machine indicates are in a state other than C(running) will be skipped.
    type: bool
    default: true
  verbose_output:
    description:
      - When V(true), include all available nodes metadata (for example C(Image), C(Region), C(Size)) as a JSON object named
        C(docker_machine_node_attributes).
    type: bool
    default: true
  filters:
    version_added: 3.5.0
"""

EXAMPLES = """
---
# Minimal example
plugin: community.docker.docker_machine

---
# Example using constructed features to create a group per Docker Machine driver
# (https://docs.docker.com/machine/drivers/), for example:
#   $ docker-machine create --driver digitalocean ... mymachine
#   $ ansible-inventory -i ./path/to/docker-machine.yml --host=mymachine
#   {
#     ...
#     "digitalocean": {
#       "hosts": [
#           "mymachine"
#       ]
#     ...
#   }
plugin: community.docker.docker_machine
strict: false
keyed_groups:
  - separator: ''
    key: docker_machine_node_attributes.DriverName

---
# Example grouping hosts by Digital Machine tag
plugin: community.docker.docker_machine
strict: false
keyed_groups:
  - prefix: tag
    key: 'dm_tags'

---
# Example using compose to override the default SSH behaviour of asking the user to accept the remote host key
plugin: community.docker.docker_machine
compose:
  ansible_ssh_common_args: '"-o StrictHostKeyChecking=accept-new"'
"""

import json
import re
import subprocess
import typing as t

from ansible.errors import AnsibleError
from ansible.module_utils.common.process import get_bin_path
from ansible.module_utils.common.text.converters import to_text
from ansible.plugins.inventory import BaseInventoryPlugin, Cacheable, Constructable
from ansible.utils.display import Display
from ansible_collections.community.library_inventory_filtering_v1.plugins.plugin_utils.inventory_filter import (
    filter_host,
    parse_filters,
)

from ansible_collections.community.docker.plugins.plugin_utils._unsafe import (
    make_unsafe,
)

if t.TYPE_CHECKING:
    from ansible.inventory.data import InventoryData
    from ansible.parsing.dataloader import DataLoader

    DaemonEnv = t.Literal[
        "require", "require-silently", "optional", "optional-silently", "skip"
    ]


display = Display()


class InventoryModule(BaseInventoryPlugin, Constructable, Cacheable):
    """Host inventory parser for ansible using Docker machine as source."""

    NAME = "community.docker.docker_machine"

    docker_machine_path: str | None = None

    def _run_command(self, args: list[str]) -> str:
        if not self.docker_machine_path:
            try:
                self.docker_machine_path = get_bin_path("docker-machine")
            except ValueError as e:
                raise AnsibleError(to_text(e)) from e

        command = [self.docker_machine_path]
        command.extend(args)
        display.debug(f"Executing command {command}")
        try:
            result = subprocess.check_output(command)
        except subprocess.CalledProcessError as e:
            display.warning(
                f"Exception {type(e).__name__} caught while executing command {command}, this was the original exception: {e}"
            )
            raise e

        return to_text(result).strip()

    def _get_docker_daemon_variables(self, machine_name: str) -> list[tuple[str, str]]:
        """
        Capture settings from Docker Machine that would be needed to connect to the remote Docker daemon installed on
        the Docker Machine remote host. Note: passing '--shell=sh' is a workaround for 'Error: Unknown shell'.
        """
        try:
            env_lines = self._run_command(
                ["env", "--shell=sh", machine_name]
            ).splitlines()
        except subprocess.CalledProcessError:
            # This can happen when the machine is created but provisioning is incomplete
            return []

        # example output of docker-machine env --shell=sh:
        #   export DOCKER_TLS_VERIFY="1"
        #   export DOCKER_HOST="tcp://134.209.204.160:2376"
        #   export DOCKER_CERT_PATH="/root/.docker/machine/machines/routinator"
        #   export DOCKER_MACHINE_NAME="routinator"
        #   # Run this command to configure your shell:
        #   # eval $(docker-machine env --shell=bash routinator)

        # capture any of the DOCKER_xxx variables that were output and create Ansible host vars
        # with the same name and value but with a dm_ name prefix.
        env_vars = []
        for line in env_lines:
            match = re.search('(DOCKER_[^=]+)="([^"]+)"', line)
            if match:
                env_var_name = match.group(1)
                env_var_value = match.group(2)
                env_vars.append((env_var_name, env_var_value))

        return env_vars

    def _get_machine_names(self) -> list[str]:
        # Filter out machines that are not in the Running state as we probably cannot do anything useful actions
        # with them.
        ls_command = ["ls", "-q"]
        if self.get_option("running_required"):
            ls_command.extend(["--filter", "state=Running"])

        try:
            ls_lines = self._run_command(ls_command)
        except subprocess.CalledProcessError:
            return []

        return ls_lines.splitlines()

    def _inspect_docker_machine_host(self, node: str) -> t.Any | None:
        try:
            inspect_lines = self._run_command(["inspect", node])
        except subprocess.CalledProcessError:
            return None

        return json.loads(inspect_lines)

    def _ip_addr_docker_machine_host(self, node: str) -> t.Any | None:
        try:
            ip_addr = self._run_command(["ip", node])
        except subprocess.CalledProcessError:
            return None

        return ip_addr

    def _should_skip_host(
        self,
        machine_name: str,
        env_var_tuples: list[tuple[str, str]],
        daemon_env: DaemonEnv,
    ) -> bool:
        if not env_var_tuples:
            warning_prefix = f"Unable to fetch Docker daemon env vars from Docker Machine for host {machine_name}"
            if daemon_env in ("require", "require-silently"):
                if daemon_env == "require":
                    display.warning(f"{warning_prefix}: host will be skipped")
                return True
            if daemon_env == "optional":
                display.warning(
                    f"{warning_prefix}: host will lack dm_DOCKER_xxx variables"
                )
            # daemon_env is 'optional-silently'
        return False

    def _populate(self) -> None:
        if self.inventory is None:
            raise AssertionError("Inventory must be there")

        daemon_env: DaemonEnv = self.get_option("daemon_env")
        filters = parse_filters(self.get_option("filters"))
        try:
            for node in self._get_machine_names():
                node_attrs = self._inspect_docker_machine_host(node)
                if not node_attrs:
                    continue

                unsafe_node_attrs = make_unsafe(node_attrs)

                machine_name = unsafe_node_attrs["Driver"]["MachineName"]
                if not filter_host(self, machine_name, unsafe_node_attrs, filters):
                    continue

                # query `docker-machine env` to obtain remote Docker daemon connection settings in the form of commands
                # that could be used to set environment variables to influence a local Docker client:
                if daemon_env == "skip":
                    env_var_tuples = []
                else:
                    env_var_tuples = self._get_docker_daemon_variables(machine_name)
                    if self._should_skip_host(machine_name, env_var_tuples, daemon_env):
                        continue

                # add an entry in the inventory for this host
                self.inventory.add_host(machine_name)

                # check for valid ip address from inspect output, else explicitly use ip command to find host ip address
                # this works around an issue seen with Google Compute Platform where the IP address was not available
                # via the 'inspect' subcommand but was via the 'ip' subcomannd.
                if unsafe_node_attrs["Driver"]["IPAddress"]:
                    ip_addr = unsafe_node_attrs["Driver"]["IPAddress"]
                else:
                    ip_addr = self._ip_addr_docker_machine_host(node)

                # set standard Ansible remote host connection settings to details captured from `docker-machine`
                # see: https://docs.ansible.com/ansible/latest/user_guide/intro_inventory.html
                self.inventory.set_variable(
                    machine_name, "ansible_host", make_unsafe(ip_addr)
                )
                self.inventory.set_variable(
                    machine_name, "ansible_port", unsafe_node_attrs["Driver"]["SSHPort"]
                )
                self.inventory.set_variable(
                    machine_name, "ansible_user", unsafe_node_attrs["Driver"]["SSHUser"]
                )
                self.inventory.set_variable(
                    machine_name,
                    "ansible_ssh_private_key_file",
                    unsafe_node_attrs["Driver"]["SSHKeyPath"],
                )

                # set variables based on Docker Machine tags
                tags = unsafe_node_attrs["Driver"].get("Tags") or ""
                self.inventory.set_variable(machine_name, "dm_tags", make_unsafe(tags))

                # set variables based on Docker Machine env variables
                for kv in env_var_tuples:
                    self.inventory.set_variable(
                        machine_name, f"dm_{kv[0]}", make_unsafe(kv[1])
                    )

                if self.get_option("verbose_output"):
                    self.inventory.set_variable(
                        machine_name,
                        "docker_machine_node_attributes",
                        unsafe_node_attrs,
                    )

                # Use constructed if applicable
                strict = self.get_option("strict")

                # Composed variables
                self._set_composite_vars(
                    self.get_option("compose"),
                    unsafe_node_attrs,
                    machine_name,
                    strict=strict,
                )

                # Complex groups based on jinja2 conditionals, hosts that meet the conditional are added to group
                self._add_host_to_composed_groups(
                    self.get_option("groups"),
                    unsafe_node_attrs,
                    machine_name,
                    strict=strict,
                )

                # Create groups based on variable values and add the corresponding hosts to it
                self._add_host_to_keyed_groups(
                    self.get_option("keyed_groups"),
                    unsafe_node_attrs,
                    machine_name,
                    strict=strict,
                )

        except Exception as e:
            raise AnsibleError(
                f"Unable to fetch hosts from Docker Machine, this was the original exception: {e}"
            ) from e

    def verify_file(self, path: str) -> bool:
        """Return the possibility of a file being consumable by this plugin."""
        return super().verify_file(path) and path.endswith(
            ("docker_machine.yaml", "docker_machine.yml")
        )

    def parse(
        self,
        inventory: InventoryData,
        loader: DataLoader,
        path: str,
        cache: bool = True,
    ) -> None:
        super().parse(inventory, loader, path, cache)
        self._read_config_data(path)
        self._populate()



#!/usr/bin/python

# Copyright Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: linode
short_description: Manage instances on the Linode Public Cloud
description:
  - Manage Linode Public Cloud instances and optionally wait for it to be 'running'.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  state:
    description:
      - Indicate desired state of the resource.
    choices: [absent, active, deleted, present, restarted, started, stopped]
    default: present
    type: str
  api_key:
    description:
      - Linode API key.
      - E(LINODE_API_KEY) environment variable can be used instead.
    type: str
    required: true
  name:
    description:
      - Name to give the instance (alphanumeric, dashes, underscore).
      - To keep sanity on the Linode Web Console, name is prepended with C(LinodeID-).
    required: true
    type: str
  displaygroup:
    description:
      - Add the instance to a Display Group in Linode Manager.
    type: str
    default: ''
  linode_id:
    description:
      - Unique ID of a Linode server. This value is read-only in the sense that if you specify it on creation of a Linode
        it is not used. The Linode API generates these IDs and we can those generated value here to reference a Linode more
        specifically. This is useful for idempotency.
    aliases: [lid]
    type: int
  additional_disks:
    description:
      - List of dictionaries for creating additional disks that are added to the Linode configuration settings.
      - Dictionary takes Size, Label, Type. Size is in MB.
    type: list
    elements: dict
  alert_bwin_enabled:
    description:
      - Set status of bandwidth in alerts.
    type: bool
  alert_bwin_threshold:
    description:
      - Set threshold in MB of bandwidth in alerts.
    type: int
  alert_bwout_enabled:
    description:
      - Set status of bandwidth out alerts.
    type: bool
  alert_bwout_threshold:
    description:
      - Set threshold in MB of bandwidth out alerts.
    type: int
  alert_bwquota_enabled:
    description:
      - Set status of bandwidth quota alerts as percentage of network transfer quota.
    type: bool
  alert_bwquota_threshold:
    description:
      - Set threshold in MB of bandwidth quota alerts.
    type: int
  alert_cpu_enabled:
    description:
      - Set status of receiving CPU usage alerts.
    type: bool
  alert_cpu_threshold:
    description:
      - Set percentage threshold for receiving CPU usage alerts. Each CPU core adds 100% to total.
    type: int
  alert_diskio_enabled:
    description:
      - Set status of receiving disk IO alerts.
    type: bool
  alert_diskio_threshold:
    description:
      - Set threshold for average IO ops/sec over 2 hour period.
    type: int
  backupweeklyday:
    description:
      - Day of the week to take backups.
    type: int
  backupwindow:
    description:
      - The time window in which backups are taken.
    type: int
  plan:
    description:
      - Plan to use for the instance (Linode plan).
    type: int
  payment_term:
    description:
      - Payment term to use for the instance (payment term in months).
    default: 1
    choices: [1, 12, 24]
    type: int
  password:
    description:
      - Root password to apply to a new server (auto generated if missing).
    type: str
  private_ip:
    description:
      - Add private IPv4 address when Linode is created.
      - Default is V(false).
    type: bool
  ssh_pub_key:
    description:
      - SSH public key applied to root user.
    type: str
  swap:
    description:
      - Swap size in MB.
    default: 512
    type: int
  distribution:
    description:
      - Distribution to use for the instance (Linode Distribution).
    type: int
  datacenter:
    description:
      - Datacenter to create an instance in (Linode Datacenter).
    type: int
  kernel_id:
    description:
      - Kernel to use for the instance (Linode Kernel).
    type: int
  wait:
    description:
      - Wait for the instance to be in state V(running) before returning.
    type: bool
    default: true
  wait_timeout:
    description:
      - How long before wait gives up, in seconds.
    default: 300
    type: int
  watchdog:
    description:
      - Set status of Lassie watchdog.
    type: bool
    default: true
requirements:
  - linode-python
author:
  - Vincent Viallet (@zbal)
notes:
  - Please note, linode-python does not have python 3 support.
  - This module uses the now deprecated v3 of the Linode API.
  - Please review U(https://www.linode.com/api/linode) for determining the required parameters.
"""

EXAMPLES = r"""
- name: Create a new Linode
  community.general.linode:
    name: linode-test1
    plan: 1
    datacenter: 7
    distribution: 129
    state: present
  register: linode_creation

- name: Create a server with a private IP Address
  community.general.linode:
    module: linode
    api_key: 'longStringFromLinodeApi'
    name: linode-test1
    plan: 1
    datacenter: 2
    distribution: 99
    password: 'superSecureRootPassword'
    private_ip: true
    ssh_pub_key: 'ssh-rsa qwerty'
    swap: 768
    wait: true
    wait_timeout: 600
    state: present
  delegate_to: localhost
  register: linode_creation

- name: Fully configure new server
  community.general.linode:
    api_key: 'longStringFromLinodeApi'
    name: linode-test1
    plan: 4
    datacenter: 2
    distribution: 99
    kernel_id: 138
    password: 'superSecureRootPassword'
    private_ip: true
    ssh_pub_key: 'ssh-rsa qwerty'
    swap: 768
    wait: true
    wait_timeout: 600
    state: present
    alert_bwquota_enabled: true
    alert_bwquota_threshold: 80
    alert_bwin_enabled: true
    alert_bwin_threshold: 10
    alert_cpu_enabled: true
    alert_cpu_threshold: 210
    alert_bwout_enabled: true
    alert_bwout_threshold: 10
    alert_diskio_enabled: true
    alert_diskio_threshold: 10000
    backupweeklyday: 1
    backupwindow: 2
    displaygroup: 'test'
    additional_disks:
      - {Label: 'disk1', Size: 2500, Type: 'raw'}
      - {Label: 'newdisk', Size: 2000}
    watchdog: true
  delegate_to: localhost
  register: linode_creation

- name: Ensure a running server (create if missing)
  community.general.linode:
    api_key: 'longStringFromLinodeApi'
    name: linode-test1
    plan: 1
    datacenter: 2
    distribution: 99
    password: 'superSecureRootPassword'
    ssh_pub_key: 'ssh-rsa qwerty'
    swap: 768
    wait: true
    wait_timeout: 600
    state: present
  delegate_to: localhost
  register: linode_creation

- name: Delete a server
  community.general.linode:
    api_key: 'longStringFromLinodeApi'
    name: linode-test1
    linode_id: "{{ linode_creation.instance.id }}"
    state: absent
  delegate_to: localhost

- name: Stop a server
  community.general.linode:
    api_key: 'longStringFromLinodeApi'
    name: linode-test1
    linode_id: "{{ linode_creation.instance.id }}"
    state: stopped
  delegate_to: localhost

- name: Reboot a server
  community.general.linode:
    api_key: 'longStringFromLinodeApi'
    name: linode-test1
    linode_id: "{{ linode_creation.instance.id }}"
    state: restarted
  delegate_to: localhost
"""

import random
import string
import time
import traceback

LINODE_IMP_ERR = None
try:
    from linode import api as linode_api

    HAS_LINODE = True
except ImportError:
    LINODE_IMP_ERR = traceback.format_exc()
    HAS_LINODE = False

from ansible.module_utils.basic import AnsibleModule, env_fallback, missing_required_lib


def randompass():
    """
    Generate a long random password that comply to Linode requirements
    """
    # Linode API currently requires the following:
    # It must contain at least two of these four character classes:
    # lower case letters - upper case letters - numbers - punctuation
    # we play it safe :)

    # as of python 2.4, this reseeds the PRNG from urandom
    random.seed()
    lower = "".join(random.choice(string.ascii_lowercase) for x in range(6))
    upper = "".join(random.choice(string.ascii_uppercase) for x in range(6))
    number = "".join(random.choice(string.digits) for x in range(6))
    punct = "".join(random.choice(string.punctuation) for x in range(6))
    p = lower + upper + number + punct
    return "".join(random.sample(p, len(p)))


def getInstanceDetails(api, server):
    """
    Return the details of an instance, populating IPs, etc.
    """
    instance = {"id": server["LINODEID"], "name": server["LABEL"], "public": [], "private": []}

    # Populate with ips
    for ip in api.linode_ip_list(LinodeId=server["LINODEID"]):
        if ip["ISPUBLIC"] and "ipv4" not in instance:
            instance["ipv4"] = ip["IPADDRESS"]
            instance["fqdn"] = ip["RDNS_NAME"]
        if ip["ISPUBLIC"]:
            instance["public"].append({"ipv4": ip["IPADDRESS"], "fqdn": ip["RDNS_NAME"], "ip_id": ip["IPADDRESSID"]})
        else:
            instance["private"].append({"ipv4": ip["IPADDRESS"], "fqdn": ip["RDNS_NAME"], "ip_id": ip["IPADDRESSID"]})
    return instance


def linodeServers(
    module,
    api,
    state,
    name,
    displaygroup,
    plan,
    additional_disks,
    distribution,
    datacenter,
    kernel_id,
    linode_id,
    payment_term,
    password,
    private_ip,
    ssh_pub_key,
    swap,
    wait,
    wait_timeout,
    watchdog,
    **kwargs,
):
    instances = []
    changed = False
    new_server = False
    servers = []
    disks = []
    configs = []
    jobs = []

    # See if we can match an existing server details with the provided linode_id
    if linode_id:
        # For the moment we only consider linode_id as criteria for match
        # Later we can use more (size, name, etc.) and update existing
        servers = api.linode_list(LinodeId=linode_id)
        # Attempt to fetch details about disks and configs only if servers are
        # found with linode_id
        if servers:
            disks = api.linode_disk_list(LinodeId=linode_id)
            configs = api.linode_config_list(LinodeId=linode_id)

    # Act on the state
    if state in ("active", "present", "started"):
        # TODO: validate all the plan / distribution / datacenter are valid

        # Multi step process/validation:
        #  - need linode_id (entity)
        #  - need disk_id for linode_id - create disk from distrib
        #  - need config_id for linode_id - create config (need kernel)

        # Any create step triggers a job that need to be waited for.
        if not servers:
            for arg in (name, plan, distribution, datacenter):
                if not arg:
                    module.fail_json(msg=f"{arg} is required for {state} state")
            # Create linode entity
            new_server = True

            # Get size of all individually listed disks to subtract from Distribution disk
            used_disk_space = 0 if additional_disks is None else sum(disk["Size"] for disk in additional_disks)

            try:
                res = api.linode_create(DatacenterID=datacenter, PlanID=plan, PaymentTerm=payment_term)
                linode_id = res["LinodeID"]
                # Update linode Label to match name
                api.linode_update(LinodeId=linode_id, Label=f"{linode_id}-{name}")
                # Update Linode with Ansible configuration options
                api.linode_update(LinodeId=linode_id, LPM_DISPLAYGROUP=displaygroup, WATCHDOG=watchdog, **kwargs)
                # Save server
                servers = api.linode_list(LinodeId=linode_id)
            except Exception as e:
                module.fail_json(msg=f"{e.value[0]['ERRORMESSAGE']}")

        # Add private IP to Linode
        if private_ip:
            try:
                res = api.linode_ip_addprivate(LinodeID=linode_id)
            except Exception as e:
                module.fail_json(msg=f"{e.value[0]['ERRORMESSAGE']}", exception=traceback.format_exc())

        if not disks:
            for arg in (name, linode_id, distribution):
                if not arg:
                    module.fail_json(msg=f"{arg} is required for {state} state")
            # Create disks (1 from distrib, 1 for SWAP)
            new_server = True
            try:
                if not password:
                    # Password is required on creation, if not provided generate one
                    password = randompass()
                if not swap:
                    swap = 512
                # Create data disk
                size = servers[0]["TOTALHD"] - used_disk_space - swap

                if ssh_pub_key:
                    res = api.linode_disk_createfromdistribution(
                        LinodeId=linode_id,
                        DistributionID=distribution,
                        rootPass=password,
                        rootSSHKey=ssh_pub_key,
                        Label=f"{name} data disk (lid: {linode_id})",
                        Size=size,
                    )
                else:
                    res = api.linode_disk_createfromdistribution(
                        LinodeId=linode_id,
                        DistributionID=distribution,
                        rootPass=password,
                        Label=f"{name} data disk (lid: {linode_id})",
                        Size=size,
                    )
                jobs.append(res["JobID"])
                # Create SWAP disk
                res = api.linode_disk_create(
                    LinodeId=linode_id, Type="swap", Label=f"{name} swap disk (lid: {linode_id})", Size=swap
                )
                # Create individually listed disks at specified size
                if additional_disks:
                    for disk in additional_disks:
                        # If a disk Type is not passed in, default to ext4
                        if disk.get("Type") is None:
                            disk["Type"] = "ext4"
                        res = api.linode_disk_create(
                            LinodeID=linode_id, Label=disk["Label"], Size=disk["Size"], Type=disk["Type"]
                        )

                jobs.append(res["JobID"])
            except Exception as e:
                # TODO: destroy linode ?
                module.fail_json(msg=f"{e.value[0]['ERRORMESSAGE']}", exception=traceback.format_exc())

        if not configs:
            for arg in (name, linode_id, distribution):
                if not arg:
                    module.fail_json(msg=f"{arg} is required for {state} state")

            # Check architecture
            for distrib in api.avail_distributions():
                if distrib["DISTRIBUTIONID"] != distribution:
                    continue
                arch = "32"
                if distrib["IS64BIT"]:
                    arch = "64"
                break

            # Get latest kernel matching arch if kernel_id is not specified
            if not kernel_id:
                for kernel in api.avail_kernels():
                    if not kernel["LABEL"].startswith(f"Latest {arch}"):
                        continue
                    kernel_id = kernel["KERNELID"]
                    break

            # Get disk list
            disks_id = []
            for disk in api.linode_disk_list(LinodeId=linode_id):
                if disk["TYPE"] == "ext3":
                    disks_id.insert(0, str(disk["DISKID"]))
                    continue
                disks_id.append(str(disk["DISKID"]))
            # Trick to get the 9 items in the list
            while len(disks_id) < 9:
                disks_id.append("")
            disks_list = ",".join(disks_id)

            # Create config
            new_server = True
            try:
                api.linode_config_create(
                    LinodeId=linode_id, KernelId=kernel_id, Disklist=disks_list, Label=f"{name} config"
                )
                configs = api.linode_config_list(LinodeId=linode_id)
            except Exception as e:
                module.fail_json(msg=f"{e.value[0]['ERRORMESSAGE']}", exception=traceback.format_exc())

        # Start / Ensure servers are running
        for server in servers:
            # Refresh server state
            server = api.linode_list(LinodeId=server["LINODEID"])[0]
            # Ensure existing servers are up and running, boot if necessary
            if server["STATUS"] != 1:
                res = api.linode_boot(LinodeId=linode_id)
                jobs.append(res["JobID"])
                changed = True

            # wait here until the instances are up
            wait_timeout = time.time() + wait_timeout
            while wait and wait_timeout > time.time():
                # refresh the server details
                server = api.linode_list(LinodeId=server["LINODEID"])[0]
                # status:
                #  -2: Boot failed
                #  1: Running
                if server["STATUS"] in (-2, 1):
                    break
                time.sleep(5)
            if wait and wait_timeout <= time.time():
                # waiting took too long
                module.fail_json(msg=f"Timeout waiting on {server['LABEL']} (lid: {server['LINODEID']})")
            # Get a fresh copy of the server details
            server = api.linode_list(LinodeId=server["LINODEID"])[0]
            if server["STATUS"] == -2:
                module.fail_json(msg=f"{server['LABEL']} (lid: {server['LINODEID']}) failed to boot")
            # From now on we know the task is a success
            # Build instance report
            instance = getInstanceDetails(api, server)
            # depending on wait flag select the status
            if wait:
                instance["status"] = "Running"
            else:
                instance["status"] = "Starting"

            # Return the root password if this is a new box and no SSH key
            # has been provided
            if new_server and not ssh_pub_key:
                instance["password"] = password
            instances.append(instance)

    elif state in ("stopped",):
        if not servers:
            module.fail_json(msg=f"Server (lid: {linode_id}) not found")

        for server in servers:
            instance = getInstanceDetails(api, server)
            if server["STATUS"] != 2:
                try:
                    res = api.linode_shutdown(LinodeId=linode_id)
                except Exception as e:
                    module.fail_json(msg=f"{e.value[0]['ERRORMESSAGE']}", exception=traceback.format_exc())
                instance["status"] = "Stopping"
                changed = True
            else:
                instance["status"] = "Stopped"
            instances.append(instance)

    elif state in ("restarted",):
        if not servers:
            module.fail_json(msg=f"Server (lid: {linode_id}) not found")

        for server in servers:
            instance = getInstanceDetails(api, server)
            try:
                res = api.linode_reboot(LinodeId=server["LINODEID"])
            except Exception as e:
                module.fail_json(msg=f"{e.value[0]['ERRORMESSAGE']}", exception=traceback.format_exc())
            instance["status"] = "Restarting"
            changed = True
            instances.append(instance)

    elif state in ("absent", "deleted"):
        for server in servers:
            instance = getInstanceDetails(api, server)
            try:
                api.linode_delete(LinodeId=server["LINODEID"], skipChecks=True)
            except Exception as e:
                module.fail_json(msg=f"{e.value[0]['ERRORMESSAGE']}", exception=traceback.format_exc())
            instance["status"] = "Deleting"
            changed = True
            instances.append(instance)

    # Ease parsing if only 1 instance
    if len(instances) == 1:
        module.exit_json(changed=changed, instance=instances[0])

    module.exit_json(changed=changed, instances=instances)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(
                type="str",
                default="present",
                choices=["absent", "active", "deleted", "present", "restarted", "started", "stopped"],
            ),
            api_key=dict(type="str", no_log=True, required=True, fallback=(env_fallback, ["LINODE_API_KEY"])),
            name=dict(type="str", required=True),
            alert_bwin_enabled=dict(type="bool"),
            alert_bwin_threshold=dict(type="int"),
            alert_bwout_enabled=dict(type="bool"),
            alert_bwout_threshold=dict(type="int"),
            alert_bwquota_enabled=dict(type="bool"),
            alert_bwquota_threshold=dict(type="int"),
            alert_cpu_enabled=dict(type="bool"),
            alert_cpu_threshold=dict(type="int"),
            alert_diskio_enabled=dict(type="bool"),
            alert_diskio_threshold=dict(type="int"),
            backupweeklyday=dict(type="int"),
            backupwindow=dict(type="int"),
            displaygroup=dict(type="str", default=""),
            plan=dict(type="int"),
            additional_disks=dict(type="list", elements="dict"),
            distribution=dict(type="int"),
            datacenter=dict(type="int"),
            kernel_id=dict(type="int"),
            linode_id=dict(type="int", aliases=["lid"]),
            payment_term=dict(type="int", default=1, choices=[1, 12, 24]),
            password=dict(type="str", no_log=True),
            private_ip=dict(type="bool"),
            ssh_pub_key=dict(type="str"),
            swap=dict(type="int", default=512),
            wait=dict(type="bool", default=True),
            wait_timeout=dict(type="int", default=300),
            watchdog=dict(type="bool", default=True),
        ),
        required_if=[
            ("state", "restarted", ["linode_id"]),
            ("state", "stopped", ["linode_id"]),
        ],
    )

    if not HAS_LINODE:
        module.fail_json(msg=missing_required_lib("linode-python"), exception=LINODE_IMP_ERR)

    state = module.params.get("state")
    api_key = module.params.get("api_key")
    name = module.params.get("name")
    alert_bwin_enabled = module.params.get("alert_bwin_enabled")
    alert_bwin_threshold = module.params.get("alert_bwin_threshold")
    alert_bwout_enabled = module.params.get("alert_bwout_enabled")
    alert_bwout_threshold = module.params.get("alert_bwout_threshold")
    alert_bwquota_enabled = module.params.get("alert_bwquota_enabled")
    alert_bwquota_threshold = module.params.get("alert_bwquota_threshold")
    alert_cpu_enabled = module.params.get("alert_cpu_enabled")
    alert_cpu_threshold = module.params.get("alert_cpu_threshold")
    alert_diskio_enabled = module.params.get("alert_diskio_enabled")
    alert_diskio_threshold = module.params.get("alert_diskio_threshold")
    backupweeklyday = module.params.get("backupweeklyday")
    backupwindow = module.params.get("backupwindow")
    displaygroup = module.params.get("displaygroup")
    plan = module.params.get("plan")
    additional_disks = module.params.get("additional_disks")
    distribution = module.params.get("distribution")
    datacenter = module.params.get("datacenter")
    kernel_id = module.params.get("kernel_id")
    linode_id = module.params.get("linode_id")
    payment_term = module.params.get("payment_term")
    password = module.params.get("password")
    private_ip = module.params.get("private_ip")
    ssh_pub_key = module.params.get("ssh_pub_key")
    swap = module.params.get("swap")
    wait = module.params.get("wait")
    wait_timeout = module.params.get("wait_timeout")
    watchdog = int(module.params.get("watchdog"))

    check_items = dict(
        alert_bwin_enabled=alert_bwin_enabled,
        alert_bwin_threshold=alert_bwin_threshold,
        alert_bwout_enabled=alert_bwout_enabled,
        alert_bwout_threshold=alert_bwout_threshold,
        alert_bwquota_enabled=alert_bwquota_enabled,
        alert_bwquota_threshold=alert_bwquota_threshold,
        alert_cpu_enabled=alert_cpu_enabled,
        alert_cpu_threshold=alert_cpu_threshold,
        alert_diskio_enabled=alert_diskio_enabled,
        alert_diskio_threshold=alert_diskio_threshold,
        backupweeklyday=backupweeklyday,
        backupwindow=backupwindow,
    )

    kwargs = {k: v for k, v in check_items.items() if v is not None}

    # setup the auth
    try:
        api = linode_api.Api(api_key)
        api.test_echo()
    except Exception as e:
        module.fail_json(msg=f"{e.value[0]['ERRORMESSAGE']}", exception=traceback.format_exc())

    linodeServers(
        module,
        api,
        state,
        name,
        displaygroup,
        plan,
        additional_disks,
        distribution,
        datacenter,
        kernel_id,
        linode_id,
        payment_term,
        password,
        private_ip,
        ssh_pub_key,
        swap,
        wait,
        wait_timeout,
        watchdog,
        **kwargs,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2013, Balazs Pocze <banyek@gawker.com>
# Copyright: (c) 2019, Andrew Klychkov (@Andersson007) <andrew.a.klychkov@gmail.com>
# Certain parts are taken from Mark Theunissen's mysqldb module
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = r'''
---
module: mysql_replication
short_description: Manage MySQL or MariaDB replication
description:
- Manages MySQL or MariaDB server replication, replica, primary status, get and change primary host.
author:
- Balazs Pocze (@banyek)
- Andrew Klychkov (@Andersson007)
- Dennis Urtubia (@dennisurtubia)
- Laurent Indermühle (@laurent-indermuehle)
options:
  mode:
    description:
    - Module operating mode. Could be
      C(changeprimary) (CHANGE MASTER TO) - also works for MySQL 8.0.23 and later since ansible.mysql 3.10.0,
      C(changereplication) (CHANGE REPLICATION SOURCE TO) - only supported in MySQL 8.0.23 and later,
      C(getprimary) (SHOW MASTER STATUS),
      C(getreplica) (SHOW REPLICA STATUS),
      C(startreplica) (START REPLICA),
      C(stopreplica) (STOP REPLICA),
      C(resetprimary) (RESET MASTER) - supported since ansible.mysql 0.1.0,
      C(resetreplica) (RESET REPLICA),
      C(resetreplicaall) (RESET REPLICA ALL).
    type: str
    choices:
    - changeprimary
    - changereplication
    - getprimary
    - getreplica
    - startreplica
    - stopreplica
    - resetprimary
    - resetreplica
    - resetreplicaall
    default: getreplica
  primary_host:
    description:
    - Same as the C(MASTER_HOST) mysql variable.
    type: str
    aliases: [master_host]
  primary_user:
    description:
    - Same as the C(MASTER_USER) mysql variable.
    type: str
    aliases: [master_user]
  primary_password:
    description:
    - Same as the C(MASTER_PASSWORD) mysql variable.
    type: str
    aliases: [master_password]
  primary_port:
    description:
    - Same as the C(MASTER_PORT) mysql variable.
    type: int
    aliases: [master_port]
  primary_connect_retry:
    description:
    - Same as the C(MASTER_CONNECT_RETRY) mysql variable.
    type: int
    aliases: [master_connect_retry]
  primary_log_file:
    description:
    - Same as the C(MASTER_LOG_FILE) mysql variable.
    type: str
    aliases: [master_log_file]
  primary_log_pos:
    description:
    - Same as the C(MASTER_LOG_POS) mysql variable.
    type: int
    aliases: [master_log_pos]
  relay_log_file:
    description:
    - Same as mysql variable.
    type: str
  relay_log_pos:
    description:
    - Same as mysql variable.
    type: int
  primary_ssl:
    description:
    - Same as the C(MASTER_SSL) mysql variable.
    - When setting it to C(yes), the connection attempt only succeeds
      if an encrypted connection can be established.
    - For details, refer to
      L(MySQL encrypted replication documentation,https://dev.mysql.com/doc/refman/8.0/en/replication-solutions-encrypted-connections.html).
    - The default is C(false).
    type: bool
    aliases: [master_ssl]
  primary_ssl_ca:
    description:
    - Same as the C(MASTER_SSL_CA) mysql variable.
    - For details, refer to
      L(MySQL encrypted replication documentation,https://dev.mysql.com/doc/refman/8.0/en/replication-solutions-encrypted-connections.html).
    type: str
    aliases: [master_ssl_ca]
  primary_ssl_capath:
    description:
    - Same as the C(MASTER_SSL_CAPATH) mysql variable.
    - For details, refer to
      L(MySQL encrypted replication documentation,https://dev.mysql.com/doc/refman/8.0/en/replication-solutions-encrypted-connections.html).
    type: str
    aliases: [master_ssl_capath]
  primary_ssl_cert:
    description:
    - Same as the C(MASTER_SSL_CERT) mysql variable.
    - For details, refer to
      L(MySQL encrypted replication documentation,https://dev.mysql.com/doc/refman/8.0/en/replication-solutions-encrypted-connections.html).
    type: str
    aliases: [master_ssl_cert]
  primary_ssl_key:
    description:
    - Same as the C(MASTER_SSL_KEY) mysql variable.
    - For details, refer to
      L(MySQL encrypted replication documentation,https://dev.mysql.com/doc/refman/8.0/en/replication-solutions-encrypted-connections.html).
    type: str
    aliases: [master_ssl_key]
  primary_ssl_cipher:
    description:
    - Same as the C(MASTER_SSL_CIPHER) mysql variable.
    - Specifies a colon-separated list of one or more ciphers permitted by the replica for the replication connection.
    - For details, refer to
      L(MySQL encrypted replication documentation,https://dev.mysql.com/doc/refman/8.0/en/replication-solutions-encrypted-connections.html).
    type: str
    aliases: [master_ssl_cipher]
  primary_ssl_verify_server_cert:
    description:
    - Same as C(MASTER_SSL_VERIFY_SERVER_CERT) MySQL/MariaDB variable.
    - The module switch automatically to C(SOURCE_SSL_VERIFY_SERVER_CERT) for MySQL 8.0.23 and later.
    - Prior to ansible.mysql 3.14.0 C(false) had no effect.
    type: bool
    version_added: '3.5.0'
  primary_auto_position:
    description:
    - Whether the host uses GTID based replication or not.
    - Same as the C(MASTER_AUTO_POSITION) mysql variable.
    type: bool
    default: false
    aliases: [master_auto_position]
  primary_use_gtid:
    description:
    - Configures the replica to use the MariaDB Global Transaction ID.
    - C(disabled) equals MASTER_USE_GTID=no command.
    - To find information about available values see
      U(https://mariadb.com/kb/en/library/change-master-to/#master_use_gtid).
    - Available since MariaDB 10.0.2.
    choices: [current_pos, replica_pos, disabled]
    type: str
    version_added: '0.1.0'
    aliases: [master_use_gtid]
  primary_delay:
    description:
    - Time lag behind the primary's state (in seconds).
    - Same as the C(MASTER_DELAY) mysql variable.
    - Available from MySQL 5.6.
    - For more information see U(https://dev.mysql.com/doc/refman/8.0/en/replication-delayed.html).
    type: int
    version_added: '0.1.0'
    aliases: [master_delay]
  connection_name:
    description:
    - Name of the primary connection.
    - Supported from MariaDB 10.0.1.
    - Mutually exclusive with I(channel).
    - For more information see U(https://mariadb.com/kb/en/library/multi-source-replication/).
    type: str
    version_added: '0.1.0'
  channel:
    description:
    - Name of replication channel.
    - Multi-source replication is supported from MySQL 5.7.
    - Mutually exclusive with I(connection_name).
    - For more information see U(https://dev.mysql.com/doc/refman/8.0/en/replication-multi-source.html).
    type: str
    version_added: '0.1.0'
  fail_on_error:
    description:
    - Fails on error when calling mysql.
    type: bool
    default: false
    version_added: '0.1.0'

notes:
   - Compatible with MariaDB or MySQL.
   - If an empty value for the parameter of string type is needed, use an empty string.

attributes:
  check_mode:
    support: none
  idempotent:
    support: partial
    details:
      - The module is not idempotent for O(mode=resetprimary), O(mode=resetreplica), and O(mode=resetreplicaall).

extends_documentation_fragment:
- ansible.mysql.mysql

seealso:
- module: ansible.mysql.mysql_info
- name: MySQL replication reference
  description: Complete reference of the MySQL replication documentation.
  link: https://dev.mysql.com/doc/refman/8.0/en/replication.html
- name: MySQL encrypted replication reference.
  description: Setting up MySQL replication to use encrypted connection.
  link: https://dev.mysql.com/doc/refman/8.0/en/replication-solutions-encrypted-connections.html
- name: MariaDB replication reference
  description: Complete reference of the MariaDB replication documentation.
  link: https://mariadb.com/kb/en/library/setting-up-replication/
'''

EXAMPLES = r'''
# If you encounter the "Please explicitly state intended protocol" error,
# use the login_unix_socket argument
- name: Stop mysql replica thread
  ansible.mysql.mysql_replication:
    mode: stopreplica
    login_unix_socket: /run/mysqld/mysqld.sock

- name: Get primary binlog file name and binlog position
  ansible.mysql.mysql_replication:
    mode: getprimary

- name: Change primary to primary server 192.0.2.1 and use binary log 'mysql-bin.000009' with position 4578
  ansible.mysql.mysql_replication:
    mode: changeprimary
    primary_host: 192.0.2.1
    primary_log_file: mysql-bin.000009
    primary_log_pos: 4578

- name: Change replication source to replica server 192.0.2.1 and use binary log 'mysql-bin.000009' with position 4578
  ansible.mysql.mysql_replication:
    mode: changereplication
    primary_host: 192.0.2.1
    primary_log_file: mysql-bin.000009
    primary_log_pos: 4578

- name: Check replica status using port 3308
  ansible.mysql.mysql_replication:
    mode: getreplica
    login_host: ansible.example.com
    login_port: 3308

- name: On MariaDB change primary to use GTID current_pos
  ansible.mysql.mysql_replication:
    mode: changeprimary
    primary_use_gtid: current_pos

- name: Change primary to use replication delay 3600 seconds
  ansible.mysql.mysql_replication:
    mode: changeprimary
    primary_host: 192.0.2.1
    primary_delay: 3600

- name: Start MariaDB replica with connection name primary-1
  ansible.mysql.mysql_replication:
    mode: startreplica
    connection_name: primary-1

- name: Stop replication in channel primary-1
  ansible.mysql.mysql_replication:
    mode: stopreplica
    channel: primary-1

- name: >
    Run RESET MASTER command which will delete all existing binary log files
    and reset the binary log index file on the primary
  ansible.mysql.mysql_replication:
    mode: resetprimary

- name: Run start replica and fail the task on errors
  ansible.mysql.mysql_replication:
    mode: startreplica
    connection_name: primary-1
    fail_on_error: true

- name: Change primary and fail on error (like when replica thread is running)
  ansible.mysql.mysql_replication:
    mode: changeprimary
    fail_on_error: true
'''

RETURN = r'''
queries:
  description: List of executed queries which modified DB's state.
  returned: always
  type: list
  sample: ["CHANGE MASTER TO MASTER_HOST='primary2.example.com',MASTER_PORT=3306"]
  version_added: '0.1.0'
'''

import os
import warnings

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.ansible.mysql.plugins.module_utils.command_resolver import (
    CommandResolver
)
from ansible_collections.ansible.mysql.plugins.module_utils.mysql import (
    get_server_version,
    get_server_implementation,
    mysql_connect,
    mysql_driver,
    mysql_driver_fail_msg,
    mysql_common_argument_spec,
)
from ansible.module_utils.common.text.converters import to_native

executed_queries = []


def get_primary_status(cursor, command_resolver):
    query = command_resolver.resolve_command("SHOW MASTER STATUS")
    cursor.execute(query)

    primarystatus = cursor.fetchone()
    return primarystatus


def get_replica_status(cursor, connection_name='', channel='', term='REPLICA'):
    if connection_name:
        query = "SHOW %s '%s' STATUS" % (term, connection_name)
    else:
        query = "SHOW %s STATUS" % term

    if channel:
        query += " FOR CHANNEL '%s'" % channel

    cursor.execute(query)
    replica_status = cursor.fetchone()
    return replica_status


def stop_replica(module, cursor, connection_name='', channel='', fail_on_error=False, term='REPLICA'):
    if connection_name:
        query = "STOP %s '%s'" % (term, connection_name)
    else:
        query = 'STOP %s' % term

    if channel:
        query += " FOR CHANNEL '%s'" % channel

    try:
        executed_queries.append(query)
        cursor.execute(query)
        stopped = True
    except mysql_driver.Warning as e:
        stopped = False
    except Exception as e:
        if fail_on_error:
            module.fail_json(msg="STOP REPLICA failed: %s" % to_native(e))
        stopped = False
    return stopped


def reset_replica(module, cursor, connection_name='', channel='', fail_on_error=False, term='REPLICA'):
    if connection_name:
        query = "RESET %s '%s'" % (term, connection_name)
    else:
        query = 'RESET %s' % term

    if channel:
        query += " FOR CHANNEL '%s'" % channel

    try:
        executed_queries.append(query)
        cursor.execute(query)
        reset = True
    except mysql_driver.Warning as e:
        reset = False
    except Exception as e:
        if fail_on_error:
            module.fail_json(msg="RESET REPLICA failed: %s" % to_native(e))
        reset = False
    return reset


def reset_replica_all(module, cursor, connection_name='', channel='', fail_on_error=False, term='REPLICA'):
    if connection_name:
        query = "RESET %s '%s' ALL" % (term, connection_name)
    else:
        query = 'RESET %s ALL' % term

    if channel:
        query += " FOR CHANNEL '%s'" % channel

    try:
        executed_queries.append(query)
        cursor.execute(query)
        reset = True
    except mysql_driver.Warning as e:
        reset = False
    except Exception as e:
        if fail_on_error:
            module.fail_json(msg="RESET REPLICA ALL failed: %s" % to_native(e))
        reset = False
    return reset


def reset_primary(module, cursor, command_resolver, fail_on_error=False):
    query = command_resolver.resolve_command('RESET MASTER')
    try:
        executed_queries.append(query)
        cursor.execute(query)
        reset = True
    except mysql_driver.Warning as e:
        reset = False
    except Exception as e:
        if fail_on_error:
            module.fail_json(msg="%s failed: %s" % (command_resolver.resolve_command('RESET MASTER'), to_native(e)))
        reset = False
    return reset


def start_replica(module, cursor, connection_name='', channel='', fail_on_error=False, term='REPLICA'):
    if connection_name:
        query = "START %s '%s'" % (term, connection_name)
    else:
        query = 'START %s' % term

    if channel:
        query += " FOR CHANNEL '%s'" % channel

    try:
        executed_queries.append(query)
        cursor.execute(query)
        started = True
    except mysql_driver.Warning as e:
        started = False
    except Exception as e:
        if fail_on_error:
            module.fail_json(msg="START REPLICA failed: %s" % to_native(e))
        started = False
    return started


def changeprimary(cursor, command_resolver, chm, connection_name='', channel=''):
    query_head = command_resolver.resolve_command("CHANGE MASTER")
    if connection_name:
        query = "%s '%s' TO %s" % (query_head, connection_name, ','.join(chm))
    else:
        query = '%s TO %s' % (query_head, ','.join(chm))

    if channel:
        query += " FOR CHANNEL '%s'" % channel

    executed_queries.append(query)
    cursor.execute(query)


def changereplication(cursor, chm, channel=''):
    query = 'CHANGE REPLICATION SOURCE TO %s' % ','.join(chm)

    if channel:
        query += " FOR CHANNEL '%s'" % channel

    executed_queries.append(query)
    cursor.execute(query)


def main():
    argument_spec = mysql_common_argument_spec()
    argument_spec.update(
        mode=dict(type='str', default='getreplica', choices=[
            'getprimary',
            'getreplica',
            'changeprimary',
            'stopreplica',
            'startreplica',
            'resetprimary',
            'resetreplica',
            'resetreplicaall',
            'changereplication']),
        primary_auto_position=dict(type='bool', default=False, aliases=['master_auto_position']),
        primary_host=dict(type='str', aliases=['master_host']),
        primary_user=dict(type='str', aliases=['master_user']),
        primary_password=dict(type='str', no_log=True, aliases=['master_password']),
        primary_port=dict(type='int', aliases=['master_port']),
        primary_connect_retry=dict(type='int', aliases=['master_connect_retry']),
        primary_log_file=dict(type='str', aliases=['master_log_file']),
        primary_log_pos=dict(type='int', aliases=['master_log_pos']),
        relay_log_file=dict(type='str'),
        relay_log_pos=dict(type='int'),
        primary_ssl=dict(type='bool', aliases=['master_ssl']),
        primary_ssl_ca=dict(type='str', aliases=['master_ssl_ca']),
        primary_ssl_capath=dict(type='str', aliases=['master_ssl_capath']),
        primary_ssl_cert=dict(type='str', aliases=['master_ssl_cert']),
        primary_ssl_key=dict(type='str', no_log=False, aliases=['master_ssl_key']),
        primary_ssl_cipher=dict(type='str', aliases=['master_ssl_cipher']),
        primary_ssl_verify_server_cert=dict(type='bool'),
        primary_use_gtid=dict(type='str', choices=[
            'current_pos', 'replica_pos', 'disabled'], aliases=['master_use_gtid']),
        primary_delay=dict(type='int', aliases=['master_delay']),
        connection_name=dict(type='str'),
        channel=dict(type='str'),
        fail_on_error=dict(type='bool', default=False),
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ['connection_name', 'channel']
        ],
    )
    mode = module.params["mode"]
    primary_host = module.params["primary_host"]
    primary_user = module.params["primary_user"]
    primary_password = module.params["primary_password"]
    primary_port = module.params["primary_port"]
    primary_connect_retry = module.params["primary_connect_retry"]
    primary_log_file = module.params["primary_log_file"]
    primary_log_pos = module.params["primary_log_pos"]
    relay_log_file = module.params["relay_log_file"]
    relay_log_pos = module.params["relay_log_pos"]
    primary_ssl = module.params["primary_ssl"]
    primary_ssl_ca = module.params["primary_ssl_ca"]
    primary_ssl_capath = module.params["primary_ssl_capath"]
    primary_ssl_cert = module.params["primary_ssl_cert"]
    primary_ssl_key = module.params["primary_ssl_key"]
    primary_ssl_cipher = module.params["primary_ssl_cipher"]
    primary_ssl_verify_server_cert = module.params["primary_ssl_verify_server_cert"]
    primary_auto_position = module.params["primary_auto_position"]
    ssl_cert = module.params["client_cert"]
    ssl_key = module.params["client_key"]
    ssl_ca = module.params["ca_cert"]
    check_hostname = module.params["check_hostname"]
    connect_timeout = module.params['connect_timeout']
    config_file = module.params['config_file']
    primary_delay = module.params['primary_delay']
    if module.params.get("primary_use_gtid") == 'disabled':
        primary_use_gtid = 'no'
    else:
        primary_use_gtid = module.params["primary_use_gtid"]
    connection_name = module.params["connection_name"]
    channel = module.params['channel']
    fail_on_error = module.params['fail_on_error']

    if mysql_driver is None:
        module.fail_json(msg=mysql_driver_fail_msg)
    else:
        warnings.filterwarnings('error', category=mysql_driver.Warning)

    login_password = module.params["login_password"]
    login_user = module.params["login_user"]

    try:
        cursor, db_conn = mysql_connect(module, login_user, login_password, config_file,
                                        ssl_cert, ssl_key, ssl_ca, None, cursor_class='DictCursor',
                                        connect_timeout=connect_timeout, check_hostname=check_hostname)
    except Exception as e:
        if os.path.exists(config_file):
            module.fail_json(msg="unable to connect to database, check login_user and "
                                 "login_password are correct or %s has the credentials. "
                                 "Exception message: %s" % (config_file, to_native(e)))
        else:
            module.fail_json(msg="unable to find %s. Exception message: %s" % (config_file, to_native(e)))

    server_version = get_server_version(cursor)
    server_implementation = get_server_implementation(cursor)
    command_resolver = CommandResolver(server_implementation, server_version)
    cursor.execute("SELECT VERSION()")
    if server_implementation == 'mariadb':
        from ansible_collections.ansible.mysql.plugins.module_utils.implementations.mariadb import replication as impl
    else:
        from ansible_collections.ansible.mysql.plugins.module_utils.implementations.mysql import replication as impl

    # Since MySQL 8.0.22 and MariaDB 10.5.1,
    # "REPLICA" must be used instead of "SLAVE"
    if impl.uses_replica_terminology(cursor):
        replica_term = 'REPLICA'
    else:
        replica_term = 'SLAVE'
        if primary_use_gtid == 'replica_pos':
            primary_use_gtid = 'slave_pos'

    if mode == 'getprimary':
        status = get_primary_status(cursor, command_resolver)
        if status and "File" in status and "Position" in status:
            status['Is_Primary'] = True
        else:
            status = dict(
                Is_Primary=False,
                msg="Server is not configured as mysql primary. "
                    "Meaning: Binary logs are disabled")

        module.exit_json(queries=executed_queries, **status)

    elif mode == "getreplica":
        status = get_replica_status(cursor, connection_name, channel, replica_term)
        # MySQL 8.0 uses Replica_...
        # MariaDB 10.6 uses Slave_...
        if status and (
                "Slave_IO_Running" in status or
                "Replica_IO_Running" in status):
            status['Is_Replica'] = True
        else:
            status = dict(Is_Replica=False, msg="Server is not configured as mysql replica")

        module.exit_json(queries=executed_queries, **status)

    elif mode == 'changeprimary':
        chm = []
        result = {}
        if primary_host is not None:
            chm.append("%s='%s'" % (command_resolver.resolve_command('MASTER_HOST'), primary_host))
        if primary_user is not None:
            chm.append("%s='%s'" % (command_resolver.resolve_command('MASTER_USER'), primary_user))
        if primary_password is not None:
            chm.append("%s='%s'" % (command_resolver.resolve_command('MASTER_PASSWORD'), primary_password))
        if primary_port is not None:
            chm.append("%s=%s" % (command_resolver.resolve_command('MASTER_PORT'), primary_port))
        if primary_connect_retry is not None:
            chm.append("%s=%s" % (command_resolver.resolve_command('MASTER_CONNECT_RETRY'), primary_connect_retry))
        if primary_log_file is not None:
            chm.append("%s='%s'" % (command_resolver.resolve_command('MASTER_LOG_FILE'), primary_log_file))
        if primary_log_pos is not None:
            chm.append("%s=%s" % (command_resolver.resolve_command('MASTER_LOG_POS'), primary_log_pos))
        if primary_delay is not None:
            chm.append("%s=%s" % (command_resolver.resolve_command('MASTER_DELAY'), primary_delay))
        if relay_log_file is not None:
            chm.append("RELAY_LOG_FILE='%s'" % relay_log_file)
        if relay_log_pos is not None:
            chm.append("RELAY_LOG_POS=%s" % relay_log_pos)
        if primary_ssl is not None:
            if primary_ssl:
                chm.append("%s=1" % command_resolver.resolve_command('MASTER_SSL'))
            else:
                chm.append("%s=0" % command_resolver.resolve_command('MASTER_SSL'))
        if primary_ssl_ca is not None:
            chm.append("%s='%s'" % (command_resolver.resolve_command('MASTER_SSL_CA'), primary_ssl_ca))
        if primary_ssl_capath is not None:
            chm.append("%s='%s'" % (command_resolver.resolve_command('MASTER_SSL_CAPATH'), primary_ssl_capath))
        if primary_ssl_cert is not None:
            chm.append("%s='%s'" % (command_resolver.resolve_command('MASTER_SSL_CERT'), primary_ssl_cert))
        if primary_ssl_key is not None:
            chm.append("%s='%s'" % (command_resolver.resolve_command('MASTER_SSL_KEY'), primary_ssl_key))
        if primary_ssl_cipher is not None:
            chm.append("%s='%s'" % (command_resolver.resolve_command('MASTER_SSL_CIPHER'), primary_ssl_cipher))
        if primary_ssl_verify_server_cert is not None:
            if primary_ssl_verify_server_cert:
                chm.append("%s=1" % command_resolver.resolve_command('MASTER_SSL_VERIFY_SERVER_CERT'))
            else:
                chm.append("%s=0" % command_resolver.resolve_command('MASTER_SSL_VERIFY_SERVER_CERT'))
        if primary_auto_position:
            chm.append("%s=1" % command_resolver.resolve_command('MASTER_AUTO_POSITION'))
        if primary_use_gtid is not None:
            chm.append("MASTER_USE_GTID=%s" % primary_use_gtid)  # MariaDB only
        try:
            changeprimary(cursor, command_resolver, chm, connection_name, channel)
        except mysql_driver.Warning as e:
            result['warning'] = to_native(e)
        except Exception as e:
            module.fail_json(msg='%s. Query == %s TO %s' % (to_native(e), command_resolver.resolve_command('CHANGE MASTER'), chm))
        result['changed'] = True
        module.exit_json(queries=executed_queries, **result)
    elif mode == "startreplica":
        started = start_replica(module, cursor, connection_name, channel, fail_on_error, replica_term)
        if started is True:
            module.exit_json(msg="Replica started ", changed=True, queries=executed_queries)
        else:
            module.exit_json(msg="Replica already started (Or cannot be started)", changed=False, queries=executed_queries)
    elif mode == "stopreplica":
        stopped = stop_replica(module, cursor, connection_name, channel, fail_on_error, replica_term)
        if stopped is True:
            module.exit_json(msg="Replica stopped", changed=True, queries=executed_queries)
        else:
            module.exit_json(msg="Replica already stopped", changed=False, queries=executed_queries)
    elif mode == 'resetprimary':
        reset = reset_primary(module, cursor, command_resolver, fail_on_error)
        if reset is True:
            module.exit_json(msg="Primary reset", changed=True, queries=executed_queries)
        else:
            module.exit_json(msg="Primary already reset", changed=False, queries=executed_queries)
    elif mode == "resetreplica":
        reset = reset_replica(module, cursor, connection_name, channel, fail_on_error, replica_term)
        if reset is True:
            module.exit_json(msg="Replica reset", changed=True, queries=executed_queries)
        else:
            module.exit_json(msg="Replica already reset", changed=False, queries=executed_queries)
    elif mode == "resetreplicaall":
        reset = reset_replica_all(module, cursor, connection_name, channel, fail_on_error, replica_term)
        if reset is True:
            module.exit_json(msg="Replica reset", changed=True, queries=executed_queries)
        else:
            module.exit_json(msg="Replica already reset", changed=False, queries=executed_queries)
    elif mode == 'changereplication':
        chm = []
        result = {}
        if primary_host is not None:
            chm.append("SOURCE_HOST='%s'" % primary_host)
        if primary_user is not None:
            chm.append("SOURCE_USER='%s'" % primary_user)
        if primary_password is not None:
            chm.append("SOURCE_PASSWORD='%s'" % primary_password)
        if primary_port is not None:
            chm.append("SOURCE_PORT=%s" % primary_port)
        if primary_connect_retry is not None:
            chm.append("SOURCE_CONNECT_RETRY=%s" % primary_connect_retry)
        if primary_log_file is not None:
            chm.append("SOURCE_LOG_FILE='%s'" % primary_log_file)
        if primary_log_pos is not None:
            chm.append("SOURCE_LOG_POS=%s" % primary_log_pos)
        if primary_delay is not None:
            chm.append("SOURCE_DELAY=%s" % primary_delay)
        if relay_log_file is not None:
            chm.append("RELAY_LOG_FILE='%s'" % relay_log_file)
        if relay_log_pos is not None:
            chm.append("RELAY_LOG_POS=%s" % relay_log_pos)
        if primary_ssl is not None:
            if primary_ssl:
                chm.append("SOURCE_SSL=1")
            else:
                chm.append("SOURCE_SSL=0")
        if primary_ssl_ca is not None:
            chm.append("SOURCE_SSL_CA='%s'" % primary_ssl_ca)
        if primary_ssl_capath is not None:
            chm.append("SOURCE_SSL_CAPATH='%s'" % primary_ssl_capath)
        if primary_ssl_cert is not None:
            chm.append("SOURCE_SSL_CERT='%s'" % primary_ssl_cert)
        if primary_ssl_key is not None:
            chm.append("SOURCE_SSL_KEY='%s'" % primary_ssl_key)
        if primary_ssl_cipher is not None:
            chm.append("SOURCE_SSL_CIPHER='%s'" % primary_ssl_cipher)
        if primary_ssl_verify_server_cert is not None:
            if primary_ssl_verify_server_cert:
                chm.append("%s=1" % command_resolver.resolve_command('MASTER_SSL_VERIFY_SERVER_CERT'))
            else:
                chm.append("%s=0" % command_resolver.resolve_command('MASTER_SSL_VERIFY_SERVER_CERT'))
        if primary_auto_position:
            chm.append("SOURCE_AUTO_POSITION=1")
        try:
            changereplication(cursor, chm, channel)
        except mysql_driver.Warning as e:
            result['warning'] = to_native(e)
        except Exception as e:
            module.fail_json(msg='%s. Query == CHANGE REPLICATION SOURCE TO %s' % (to_native(e), chm))
        result['changed'] = True
        module.exit_json(queries=executed_queries, **result)

    warnings.simplefilter("ignore")


if __name__ == '__main__':
    main()
